/***********************************************************
    PROGRAM:    pictps
    MODULE:     drawelmt.c

    AUTHOR:     Patrick Hayden
    DATE:       1991-10-29

    DESCRIPTION:

	Picture interpretation code
************************************************************/
#include "pictps.h"
#include "externs.h"

extern bool pinned;

/*-----------------------------------------------------
 Import:	Pointer to an element header
 Export:	Pointer to corresponding element info
------------------------------------------------------*/

char *element_ptr ( element )
	HDR *element;
{
	char *tmp;

	tmp = (char *)element + sizeof(HDR) + element->tag[0];
	return(tmp);
#if 0
	return( (char *)element + sizeof(HDR) + element->tag[0] );
#endif
}

/*-----------------------------------------------------
 Import:	Pointer to an element header
 Export:	---

 Calls:		label_ptr() and transform_ptr()

 Points 'data'  to element extra data
 Points 'label' to element label structure
 Points 'transform' to element transformation info
------------------------------------------------------*/

ELEMENT_DATA *data_ptr( element )
	HDR *element;
{
	unsigned *length;
	char *temp;

	if ( !(element->flags & EXTRA_DATA) ) {
		label = NULL;
		transform = NULL;
		return (NULL);
	}

	/* End of element */
	temp   = (char *)element + element->length;

	/* End minus 2 bytes */
	length = (unsigned *)temp - 1;

	data	= (ELEMENT_DATA *) ( temp - *length );
	label 	= label_ptr (data);

	if (label != NULL)
		flog("Label: %s\n",label->text);

	transform = transform_ptr (data);
	return (data);
}

LABEL_DATA *label_ptr( _data_ptr )
	ELEMENT_DATA *_data_ptr;
{
	if (_data_ptr->label == 0) return (NULL);

	return( (LABEL_DATA *) ((char *)_data_ptr + _data_ptr->label) );
}

TRANSFORM_DATA *transform_ptr(_data_ptr)
	ELEMENT_DATA *_data_ptr;
{
	if (_data_ptr->transform == 0) return (NULL);

	return( (TRANSFORM_DATA *) ((char *)_data_ptr + _data_ptr->transform) );
}

int Draw_element( element )
	register HDR *element;
{
	FXY_AREA xy;
	char *tag;

	/* Meta elements don't need full processing */
	if ( element->type == META_ELEMENT ) {
    	Draw_meta( element );
	    return ( 1 );
	}

	tag = (char *)element + sizeof (HDR);

	/* BUTTON == element supplied button */
	if (!strcmp(tag,"BUTTON"))
		element->flags |= HIDDEN;

	if( element->flags & HIDDEN )
		return(0);

	/* An optimization */
	if (element->fill == TRANSPARENT)
		element->pattern = SOLID_PAT;

	/* Initialize internally used structure members */
	element->_col = element->_row = 0;
	element->_width = element->_height = 0;

	if ( flags & STAT_PRT ) {
    	/* Modify Frank's BACK element */
	    if (!strcmp(tag,"BACK") ) {
    	    element->fill = element->color = TRANSPARENT;
			if ( !pinned ) {
				element->col += PH(2);
				element->width -= PH(2);
			}
			if( wind.state.active )
				element->flags |= STATE_SET;
		}

		/* Replace this sympol with a stroked window button */
	    if ( !strcmp(tag,"WINDOW") ) {
			xy_map( (IRECT_AREA *)&element->row, &xy );
	        buf_printf( "%d %d abbr\n", xy.x, xy.y );
    	    return( 1 );
	    }
	}
	data = data_ptr(element);

	if ( !( options & OPT_MIN_EPS ) ) 
		{
		if ( *tag && isprintable(tag) && !( flags & STAT_SHADOWING ) )
    		buf_printf("%%%%BeginObject: %s\n",tag);

		else if ( isprint( element->type ) )  
    		buf_printf("%%%%BeginObject: Type%c\n",element->type);

		else
    		buf_printf("%%%%BeginObject: ControlType\n");
		}

	if( element->attr & DIMMED && !( flags & STAT_SHADOWING ) ) {
		flags |= STAT_DIMMING;
		buf_printf( "/CC1 [ 0.36 0.36 0.36 ] def\n/CC4 [ 0.52 0.52 0.52 ] def\n/CC5 [ 0.86 0.86 0.86 ] def\n" );
	} 

	Set_coords ( element );

	draw_transform_start();

	if ( flags & STAT_DISPLAYING ) {
		PictureCurrent( display_pid );
		DrawPicture( element, 0, NULL, "A" );
		Draw();
	}

	/* Modify elements according to grouping rules */
	if ( (element->type != GROUP) && (element->flags & IN_GROUP) && group_element ) {

		/* Resolve label size and alignment */
        if ( group_element->list && !group_element->scroll ) 
		 	if ( element->type == TEXT && !group_hdr[group_level]->_row ) {
				/* Fake up our group element */
				group_hdr[group_level]->type = TEXT;
				group_hdr[group_level]->_row = element->row;
				/*  Frank hasn't thought to do this
				group_hdr[group_level]->size.text.height =
			  		element->size.text.height;
				*/
				group_hdr[group_level]->size.text.height = PV(11);
		 	}
		
		/* Widen bounding boxes if necessary */
		if ( group_element->vertical && group_element->same_width ) {
			element->width = group_hdr[group_level]->width;
			element->_width = 1;
		} else if ( group_element->scroll ) {
			element->width = group_hdr[group_level]->width - element->col - PH(4);
			element->_width = 1;
		}
	}

	if( strchr( TEXT_TYPES, element->type ) ) {
		/* Do wild-wacky text-type kind of things */
		if( label ) {
			if( label->options & LABEL_OVER ) {
				/* This will not always work but it will always make sense */
				label->options |= label->options | LABEL_LEFT;
				label->options &= ~LABEL_OVER;
			}
		}
	}

	switch(element->type & 0xff) {                             	
		case ARC:               Draw_arc(element);          	break;
		case BUTTON:            Draw_button(element);       	break;
		case CURVE:             Draw_curve(element);        	break;
		case DYN_LINK:          Draw_link(element);         	break;
		case GROUP: 			Draw_group(element);        	break;

		case END_OF_GROUP:      Draw_end_group(element);    	break;

		case IMAGE:             Draw_image(element);        	break;
		case LINE:              Draw_line(element);         	break;
		case NUMBER:            Draw_number(element);       	break;
		case PARAGRAPH:         Draw_para(element);         	break;

#if 0
		case PATH:              Draw_path(element);         	break;
#endif
		case POINTS:            Draw_points(element);       	break;
		case RECTANGLE:         Draw_rect(element);         	break;
		case SYMBOL:            Draw_symbol(element);       	break;
		case TEXT:              Draw_text(element);         	break;
		default:
			buf_printf("%%Unkwon element\n");
			return 1;
	}

	draw_transform_end();
	if( element->attr & DIMMED && !( flags & STAT_SHADOWING ) ) {
		buf_printf( "/CC1 [ 0 0 0 ] def\n/CC4 [ %.3f dup dup ] def\n/CC5 [ 1 1 1 ] def\n",
			rgb_map[4].r );
		flags &= ~STAT_DIMMING;
	}
	if ( !( flags & STAT_SHADOWING ) && !( options & OPT_MIN_EPS ) )
		buf_printf("%%%%EndObject\n");
    return( 1 );
}
