/***********************************************************
    PROGRAM:    pictps
    MODULE:     drawobjs.c

    AUTHOR:     Patrick Hayden
    DATE:       1991-10-29

    DESCRIPTION:

	Picture object interpretation code
	All PostScript specific calls should gradually be
	removed from here...
************************************************************/

#include "pictps.h"
#include "externs.h"

 
/*------------------------------------------------------------
	One function for each type of element
	Arguments are a pointer to an element header followed by
		extra element data

	Before calling, 'data', 'label' and 'transform' are
	determined by 'Draw_element'.
--------------------------------------------------------------*/


void Draw_arc( element )
	HDR *element;
{
	register ARC_ELEMENT 	*arc;
	FXY_AREA 				bbox, elmt_xy;
	unsigned				r;
	float 					yscale;

	/* Disallow zero-size arcs */
	if( !element->width || !element->height ) return;

	arc = (ARC_ELEMENT *)element_ptr( element );
	xy_map( (IRECT_AREA *)&element->row, &elmt_xy );

	memmove( &bbox, &elmt_xy, sizeof(FXY_AREA) );
	draw_options( element, &bbox );

	r = elmt_xy.w / 2;
	yscale = (float)elmt_xy.h / elmt_xy.w;
	elmt_xy.x += r;

	draw_line_style_start( (LINE_ELEMENT *)arc );

//....Modif for debug the thickness arc by Guy Grimard at Multimeg.......

	if( element->size.thickness < H_TPP ) element->size.thickness = H_TPP;
	hdr_printf( 0, 0,element->size.thickness, element->color, element->fill,element->pattern );

//.......................................................................

	/* arc->start and end are longs in QNX4 */

#ifdef __WATCOMC__
	buf_printf( "%d %.2f %u %.2f %.2f %f ",elmt_xy.x, elmt_xy.y / yscale + r, r, 
		arc->end>0.0?arc->end/1000.0:360.0, arc->start/1000.0, yscale);
#elif !defined __WATCOMC__

	buf_printf( "%d %.2f %u %.2f %.2f %f ",elmt_xy.x, elmt_xy.y / yscale + r, r,  
		arc->end>0.0?arc->end:360.0,arc->start,yscale);

#endif
	
	name_printf( "A", element->pattern );

	draw_line_style_end( (LINE_ELEMENT *)arc );
	draw_label( label, element, 0 );
}

void Draw_button( element )
	HDR *element;
{
	register BUTTON_ELEMENT	*button;
	FXY_AREA 				elmt_xy;
	RECT_AREA 				bbox, ebox;
	int 					i;

	button = (BUTTON_ELEMENT *)element_ptr( element );

#if 0
	flog("Button:\n check %d\n box %d\n menu %d\n abbr %d\n std %d\n",
		button->check, button->box, button->menu, button->abbreviated, button->standard );
#endif

	/* Fake up standard format */
	if( button->standard ) {
		if( !element->color )	element->color = BLACK;
		if( !element->fill )	element->fill = BKGD1;
		element-> attr |= MODELLED;
	}

	if( element->attr & MODELLED )
		element->attr &= ~SHADOWED;

	if( button->box )
		button->roundness = 0;

	memmove( &ebox, &element->row, sizeof(RECT_AREA) );
	xy_map( (IRECT_AREA *)&element->row, &elmt_xy );

	draw_options( element, &elmt_xy );

	i = find_font( element, (TEXT_ELEMENT *)button,
	     button->face?button->string + strlen( button->string ) + 1:NULL );

	if( element->color != TRANSPARENT ) {
		if( button->round ) {
			bbox.row = element->row;
			bbox.col = element->col;
			xy_map( (IRECT_AREA *)&bbox, &elmt_xy );
			buf_printf( "%d %d round", elmt_xy.x, elmt_xy.y );
			if( element->flags & STATE_SET )
				buf_printf( "in\n" );
			else
				buf_printf( "out\n" );
			ebox.col += RADIO_W + PH(4);
			ebox.row += ( RADIO_H - font[i].map.height ) / 2 + 
				font[i].map.baseline + PV(1);
		} else if( button->check ) {
			bbox.row = element->row - PV(4);
			bbox.col = element->col;
			xy_map( (IRECT_AREA *)&bbox, &elmt_xy );
			buf_printf( "%d %d check", elmt_xy.x, elmt_xy.y );
			if( element->flags & STATE_SET )
				buf_printf( "in\n" );
			else
				buf_printf( "out\n" );
			ebox.col += CHECK_W + PH(4);
			ebox.row += ( CHECK_H - font[i].map.height ) / 2 + 
				font[i].map.baseline - PV(2);
		} else if( button->abbreviated || button->menu ) {
			bbox.row = element->row - PV(4);
			bbox.col = element->col;
			xy_map( (IRECT_AREA *)&bbox, &elmt_xy );
			buf_printf( "%d %d abbr\n", elmt_xy.x, elmt_xy.y );
			ebox.col += ABBR_W + PH(4);
			ebox.row += ( ABBR_H - font[i].map.height ) / 2 + 
				font[i].map.baseline - PV(2);
		} else if( element->attr & MODELLED ) {
			draw_rect( &elmt_xy, TRANSPARENT, element->fill, button->roundness, 
				H_TPP, element->pattern);
			xy_map( (IRECT_AREA *)&ebox, &elmt_xy );
			draw_rect_mod( &elmt_xy, MOD_OUT, button->roundness, PH(2) );
			ebox.row += PV(5) + font[i].map.baseline;
			ebox.col += PH(9); 
		} else {
			if( flags & STAT_SHADOWING )
				button->outline = BLACK;
			draw_rect( &elmt_xy, button->outline, element->fill, 
				button->roundness, H_TPP, element->pattern );
			ebox.row += PV(5) + font[i].map.baseline;
			ebox.col += PH(9);
		}

		if( element->attr & MODELLED & ENCLOSED )
			draw_rect_mod( &elmt_xy, MOD_OUT, 0, H_TPP );

		xy_map( (IRECT_AREA *)&ebox, &elmt_xy );
		draw_text( button->string, &elmt_xy, element->color,element, (TEXT_ELEMENT *)button, i );
	}
	draw_label( label, element, ebox.row );
}

void Draw_curve ( element )
	HDR *element;
{
	register CURVE_ELEMENT	*curve;
	FXY_AREA 				elmt_xy;
	int						x,y;
	unsigned 				row, col;
	int 					i, j;

	curve = (CURVE_ELEMENT *)element_ptr( element );
	buf_printf( "Gs NP\n" );
	draw_line_style_start( curve );

//..........Modif for debug curve by Guy Grimard at Multimeg............

	push_offset();
	add_offset( (RECT_AREA *)&element->row );

	get_points( curve->compact, curve->points,0, &row, &col );

	pxy_map(&x,&y,col,row);
	buf_printf( "	%d %d Mto\n", x, y );

//.......................................................................


	for( i=1; i<=curve->n/3; i++ ) 
		{
		buf_printf( "	" );
		for ( j=1; j<=3; j++ ) 
			{
			get_points( curve->compact, curve->points, (i-1)*3+j, &row, &col );
			pxy_map( &x, &y, col, row );
			buf_printf( "%d %d ",x,y);
			}
		buf_printf( "Cto\n" );
		}

	if ( element->fill )
		buf_printf( "	%s ffill\n", ps_colour( element->fill ) );

	/* PMH 1-12-91 */
	if( element->size.thickness < H_TPP ) element->size.thickness = H_TPP;
	buf_printf( "	0 0 %d %s sstroke\nGr\n", element->size.thickness,
		ps_colour( element->color ) );

	draw_line_style_end( curve ); 

	draw_options( element, &elmt_xy );
	draw_label( label, element, 0 );

	pop_offset();
#if 0
	sub_offset( (RECT_AREA *)&element->row );
#endif
}

void Draw_group( element )
	HDR *element;
{
	register GROUP_ELEMENT 	*ptr;
	FXY_AREA 				elmt_xy, scroll, tmp_xy;

	++group_level;
	if( group_level > MAX_GROUP_NEST ) 	
		return;
	
	ptr = (GROUP_ELEMENT *)element_ptr(element);

	group_hdr[group_level] = (HDR *)salloc( element->length );
	memcpy( group_hdr[group_level], element, element->length );
	group_element = (GROUP_ELEMENT *)element_ptr( group_hdr[group_level] );

	push_offset();
	add_offset( (RECT_AREA *)&element->row );
	/* Subtract offset */
	element->row = element->col = 0;

/*
	boxclip_start( &elmt_xy );  Needs a Gr!!!
*/

	/* Draw group stuff here */

	if( ptr->scroll ) {
		xy_map( (IRECT_AREA *)&element->row, &elmt_xy );

		/* Draw chiselled */
		memcpy( &tmp_xy, &elmt_xy, sizeof(FXY_AREA) );
		draw_rect_mod( &tmp_xy, MOD_OUT, 0, H_TPP );
		memcpy( &tmp_xy, &elmt_xy, sizeof(FXY_AREA) );
		tmp_xy.x -= H_TPP;
		tmp_xy.y -= V_TPP;
		tmp_xy.w += PH(2);
		tmp_xy.h += PH(2);
		draw_rect_mod( &tmp_xy, MOD_IN, 0, H_TPP );

		/* The scroll bar is included with the frame information */
		if( !( flags & STAT_FRAME ) ) {
			scroll.x = elmt_xy.x + elmt_xy.w + PH(3);
			scroll.y = elmt_xy.y;
			scroll.h = elmt_xy.h;
			draw_vscroll( &scroll );
		}
	}

	if ( ptr->scroll )
		boxclip_start ( &elmt_xy );
}

void Draw_end_group ( element )
	HDR *element;
{
	register GROUP_ELEMENT 	*ptr;

	if( group_level == -1 || element == NULL ) return;
	if( group_level > MAX_GROUP_NEST ) 
		{
		group_level--;
		return;
		}

#if 1

	ptr = (GROUP_ELEMENT *)element_ptr( group_hdr[group_level] );
	if (ptr->scroll) clip_end();

	/* Find group information in element struct */
	data_ptr( group_hdr[group_level] );	

	/* Replace origin */
	pop_offset();
	if (group_hdr[group_level]->_row) group_hdr[group_level]->_row += group_hdr[group_level]->row;

#endif

#if 0
	sub_offset( (RECT_AREA *)&group_hdr[group_level]->row);
#endif

	/* Draw the label */
	draw_label( label, group_hdr[group_level], 0 );
	
//	buf_printf( "%%....GROUPE LEVEL...%d\n",group_level); 

#if 1

	if ( group_hdr[group_level] )
		{ 
		free( group_hdr[group_level] );
		group_hdr[group_level] = NULL;
		}
		
	/* group_element reverts to next level if it exists */

	if(group_level) group_element=(GROUP_ELEMENT *)element_ptr(group_hdr[group_level-1]);
	group_level--;

//	buf_printf( "%%....GROUPE LEVEL...%d\n",group_level); 
#endif
}

void Draw_image( element )
	HDR *element;
{
	IMAGE_ELEMENT 	*image;
	FXY_AREA 		elmt_xy;

	image = ( IMAGE_ELEMENT *)element_ptr( element );
	draw_image( element, image );
	xy_map( (IRECT_AREA *)&element->row, &elmt_xy );
	draw_options( element, &elmt_xy );
	draw_label( label, element, 0 );

}

void Draw_line( element )
	HDR *element;
{
	register LINE_ELEMENT 	*line;
	FXY_AREA 				elmt_xy;
	int 					i, x, y, x1, y1;
	unsigned 				row, col;
	char 					temp[10];

	line = (LINE_ELEMENT *)element_ptr( element );
	push_offset();
	add_offset( (RECT_AREA *)&element->row );
	get_points( line->compact, line->points, 0, &row, &col );
	pxy_map( &x, &y, col, row );
	
	draw_line_style_start( line );

	if (line->n<=2)  {  /* It's a line */
		
		get_points( line->compact, line->points, 1, &row, &col );
		pxy_map( &x1, &y1, col, row );

		/* PMH 1-12-91 */
		if( element->size.thickness < H_TPP ) element->size.thickness = H_TPP;
		buf_printf("0 0 %d %s ", element->size.thickness,
			 ps_colour(element->color) );

		if( line->end_arrow ) {
			/* Frank...
				long arrowheads are wide 
				wide arrowheads are long
				... go figure!
			*/
			if ( line->long_arrow )
				strcpy( temp, "EWAline");
			else if( line->wide_arrow )
				strcpy( temp, "ELAline" );
			else if( line->indented_arrow )
				strcpy( temp, "EIAline" );
			else strcpy( temp, "EIAline" );
		} else
			strcpy( temp, "line" );

		buf_printf("%d %d %d %d %s\n", x1, y1, x, y, temp);

		if( line->start_arrow ) {
			/* Frank...
				long arrowheads are wide 
				wide arrowheads are long
				... go figure!
			*/
			if ( line->long_arrow )
				strcpy( temp, "EWAline");
			else if( line->wide_arrow )
				strcpy( temp, "ELAline" );
			else if( line->indented_arrow )
				strcpy( temp, "EIAline" );
			else strcpy( temp, "EIAline" );
			buf_printf("0 0 %d %s ", element->size.thickness,
				 ps_colour(element->color) );
         	buf_printf("%d %d %d %d %s\n", x, y, x1, y1, temp);
		}

		elmt_xy.x = x;
		elmt_xy.y = y;
		elmt_xy.w = x1 - x;
		elmt_xy.h = y1 - y;
	} else { /* It's a polygon */
		/* Init */
		buf_printf( "Gs NP\n" );
		buf_printf( "	%d %d Mto\n", x, y );

		elmt_xy.x = elmt_xy.w = x;
		elmt_xy.y = elmt_xy.h = y;

		/* Lines to each node */
		for( i=1; i<line->n; i++ ) {
			get_points( line->compact, line->points, i, &row, &col );
			pxy_map( &x, &y, col, row );

			/* Determine bounding box */
			if ( x<elmt_xy.x ) elmt_xy.x = x;
			if ( x>elmt_xy.w ) elmt_xy.w = x;
			if ( y<elmt_xy.y ) elmt_xy.y = y;
			if ( y>elmt_xy.h ) elmt_xy.h = y;

			buf_printf("	%d %d Lto\n",x,y);
		}

		if( element->fill )
			buf_printf("	%s ffill\n",ps_colour( element->fill ) );	/* Tab char */

		if( element->size.thickness < H_TPP ) element->size.thickness = H_TPP;
		buf_printf("	%d %d %d %s sstroke\n", 0, 0, element->size.thickness,
			ps_colour(element->color) );

		buf_printf("Gr\n");

		elmt_xy.w -= elmt_xy.x;
		elmt_xy.h -= elmt_xy.y;
	}
	pop_offset();
#if 0
	sub_offset( (RECT_AREA *)&element->row );
#endif

	draw_line_style_end( line );

	draw_options( element, &elmt_xy );
	draw_label( label, element, 0 );
}

void Draw_link( element )
	HDR *element;
{
	LINK_ELEMENT 	*link;

	link = (LINK_ELEMENT *)element_ptr(element);

	resolve_link( element, link );
}

void Draw_meta( element )
    HDR *element;
{
	register META_HDR       *meta;

	meta = (META_HDR *)element;

	if( meta->change ) {
		current_point.active = YES;
		if ( meta->offset ) {
    		current_point.offset.row = meta->row;
    		current_point.offset.col = meta->col;
		} else if( meta->relative ) {
			current_point.row += meta->row;
			current_point.col += meta->col;
		} else if( meta->spacing ) {
			current_point.spacing.row = meta->row;
			current_point.spacing.col = meta->col;
		} else {
			current_point.row = meta->row;
			current_point.col = meta->col;
		}
	}
}

bool Set_coords( element )
	HDR *element;
{
	/* Only active when meta elements are found */
	if( current_point.active == NO ) 
		return ( NO );

	/* Check for a recent DrawAt (Meta element) */
	if( !current_point.col && !current_point.row ) {

		/* Get the co-ords of the last drawn element */
		current_point.row = current_point.last.row;
		current_point.col = current_point.last.col;

		/* Compute offsets */
		if ( current_point.offset.row == -1 )
			current_point.row += current_point.last.height;
		else
			current_point.row += current_point.offset.row;

		if ( current_point.offset.col == -1 )
			current_point.col += current_point.last.width;
		else
			current_point.col += current_point.offset.col;

		/* Compute spacing */
		current_point.row += current_point.spacing.row;
		current_point.col += current_point.spacing.col;
	}

	element->row = current_point.row;
	element->col = current_point.col;

	/* Remember this element */
	memmove( &current_point.last, &element->row, sizeof ( RECT_AREA ) );
	current_point.row = current_point.col = 0;                               	

	return ( YES );
}
		
void Init_coords( VOID )
{
	current_point.active = NO;
	current_point.row = current_point.col = 0;
	current_point.offset.row = -1;
	current_point.offset.row = 0;
	current_point.spacing.row = current_point.spacing.row = 0;
	current_point.last.row = current_point.last.col = 0;
	current_point.last.height = current_point.last.width = 0;
}

void Draw_number( element )
	HDR *element;
{
	/* Complicated object... */
	draw_number( element );
}

void Draw_para( element	)
	HDR *element;
{
	register PARAGRAPH_ELEMENT 	*para;
	FXY_AREA 				elmt_xy, bbox_xy;
	RECT_AREA 				bbox;
	int 					i, n, pos;
	char					*p0, *p1;

	para = (PARAGRAPH_ELEMENT *)element_ptr( element );

	if( !( element->attr & ENCLOSED ) && ( element->attr & (HIGHLIGHT|SELECTED ) ) )
		draw_shadow( element, WHITE );

	xy_map( (IRECT_AREA *)&element->row, &elmt_xy );

	i = find_font( element, (TEXT_ELEMENT *)para,
		  para->face?para->string + strlen( para->string )+1:NULL );

	/* Determine text's bounding box (Add some extra room) */
	bbox.row = element->row - PV(2);
	bbox.col = element->col - PH(4);
	if( element->height )
		bbox.height = element->height + PV(2);
	else
		bbox.height = font[i].map.height + PV(2);
	/* ???? 13pt font is treated differently */
	if( i == 2 ) {
		bbox.row += PV(1);
		bbox.height -= PV(2);
	}
	bbox.width = element->width - PH(5); /* Don't ask why */

	xy_map( (IRECT_AREA *)&bbox, &bbox_xy );
	if( element->fill != TRANSPARENT && !(flags & STAT_SHADOWING) )
		draw_rect( &bbox_xy, 0, element->fill, 0, 0, 0 );

	draw_options( element, &bbox_xy );
	bbox.row = element->row + font[i].map.baseline;
	bbox.col = element->col;
	bbox.height = font[i].map.height;
	bbox.width = element->width;
	pos = bbox.col;
	for( p0 = p1 = para->string, n=0; 
	  *p1 && n < para->length; ++p1, ++n, pos += font[i].map.width ) {
		if( ( pos >= bbox.col + bbox.width ) || *p1 == '\n' ) {
			for(; p1 > p0 && *p1 != 0x20 && *p1 != '\n'; --p1, --n )
				;
			*p1 = '\0';
			xy_map( (IRECT_AREA *)&bbox, &bbox_xy );
			draw_text( p0, &bbox_xy, element->color, element, (TEXT_ELEMENT *)para, i);
			bbox.row += font[i].map.height;
			pos = bbox.col;
			p0 = ++p1;
		}
	}
	if( pos < bbox.col + bbox.width ) {
		*p1 = '\0';
		xy_map( (IRECT_AREA *)&bbox, &bbox_xy );
		draw_text( p0, &bbox_xy, element->color, element, (TEXT_ELEMENT *)para, i);
	}
	draw_label( label, element, element->row + font[i].map.baseline );
}

void Draw_path( element )
	HDR *element;
{
	if( element == NULL )
		return;
#if 0
	register PATH_ELEMENT 	*path;
	FXY_AREA 				elmt_xy;

	path = (PATH_ELEMENT *)element_ptr(element);
	xy_map ( (IRECT_AREA *)&element->row, &elmt_xy );
	draw_options(element, &elmt_xy);
#endif
}

void Draw_points( element )
	HDR *element;
{
	register POINTS_ELEMENT	*points;
	FXY_AREA 				elmt_xy;
	int 					i, fw, fh;
	unsigned 				ux, uy;
	int						x, y;
	char					buf[2];

	points = (POINTS_ELEMENT *)element_ptr( element );

	xy_map( (IRECT_AREA *)&element->row, &elmt_xy );
	boxclip_start( &elmt_xy );

	push_offset();
	add_offset( (RECT_AREA *)&element->row );

	get_points( points->compact, (POINT *)points->points, 0, &ux, &uy);
	pxy_map( &x, &y, ux, element->height - uy );

	/* Init */
	if( points->line_color ) {
		draw_line_style_start( (LINE_ELEMENT *)&points->stroke );
		buf_printf( "Gs NP\n" );
		if( drsikes )
			buf_printf( "	%d SLW\n", drsikes );
		buf_printf( "	%d %d Mto\n", x, y );
	}

	elmt_xy.x = elmt_xy.w = x;
	elmt_xy.y = elmt_xy.h = y;

	/* Lines to each node */
	for( i=1; i<points->n; i++ ) {
		get_points( points->compact, (POINT *)points->points, i, &ux, &uy );
		pxy_map( &x, &y, ux, element->height - uy );

		/* Determine bounding box */
		if( x<elmt_xy.x ) elmt_xy.x = x;
		if( x>elmt_xy.w ) elmt_xy.w = x;
		if( y<elmt_xy.y ) elmt_xy.y = y;
		if( y>elmt_xy.h ) elmt_xy.h = y;

		if (points->line_color)
			buf_printf("	%d %d Lto\n",x,y);
		else if (points->dot)
			buf_printf("%s %d %d %d SCNr\n",ps_colour(element->color),
				H_TPP, x, y );
	}

	if (points->line_color) {
		if (element->fill)
			buf_printf("	%s ffill\n",ps_colour(element->fill) );

		buf_printf("	%d %d %d %s sstroke\n", 0, 0, H_TPP,
			ps_colour(points->line_color) );

		buf_printf( "Gr\n" );
		draw_line_style_end( (LINE_ELEMENT *)&points->stroke );
	}

	if ( ( !points->no_mark ) && ( !points->dot )) {
		if ( !points->char_mark )
			points->character = '*';

		/* To center char */
		fw = font[0].map.width / 2;
		fh = font[0].map.baseline / 2;

		for ( i=0; i<points->n; i++ ) {
			get_points( points->compact, (struct point_data *)points->points,
				 i, &ux, &uy);
			pxy_map( &x, &y, ux - fw, element->height - uy + fh );
			buf[0] = (char)points->character;
			buf[1] = 0;
			buf_printf("(%s) %d %d /QWFont0 %s text\n",strformat( buf ), x, y, ps_colour(element->color) );
		}
	}

	elmt_xy.w -= elmt_xy.x;
	elmt_xy.h -= elmt_xy.y;
#if 0
	sub_offset( (RECT_AREA *)&element->row );
#endif
	pop_offset();
	clip_end();

	draw_options( element, &elmt_xy );
	draw_label( label, element, 0 );
}

#define THICKNESS element->size.thickness

void Draw_rect( element )
	HDR *element;
{
	register RECTANGLE_ELEMENT 	*rectangle;
	FXY_AREA 					elmt_xy, tmp_xy;

	rectangle = (RECTANGLE_ELEMENT *)element_ptr( element );
	xy_map( (IRECT_AREA *)&element->row, &elmt_xy );
	memcpy( &tmp_xy, &elmt_xy, sizeof(FXY_AREA) );

	if( THICKNESS < H_TPP ) THICKNESS = H_TPP;
	if( rectangle->line_format ) element->color = TRANSPARENT;


	draw_line_style_start( (LINE_ELEMENT *)rectangle );

	if( element->color == TRANSPARENT && element->fill )
		element->color = element->fill;

	draw_options( element, &elmt_xy );

	draw_rect( &tmp_xy, element->color, element->fill, rectangle->roundness,
		element->size.thickness, element->pattern );

	draw_line_style_end( (LINE_ELEMENT *)rectangle );

	switch (rectangle->line_format & 0xff) {
	case LINE_MODEL:
	case LINE_DEPRESS:
		memcpy( &tmp_xy, &elmt_xy, sizeof(FXY_AREA) );
/*
		tmp_xy.x += THICKNESS/2;
		tmp_xy.y += THICKNESS/2;
		tmp_xy.w -= THICKNESS;
		tmp_xy.h -= THICKNESS;
*/
		if( rectangle->line_format == LINE_MODEL )
			draw_rect_mod( &tmp_xy, MOD_OUT,
				rectangle->roundness, THICKNESS );
		else
			draw_rect_mod( &tmp_xy, MOD_IN, 
				rectangle->roundness, THICKNESS );
		element->color = TRANSPARENT;
		break;
	case LINE_CHISEL:
		/* Frank doesn't support round/chiselled */
		if ( rectangle->roundness ) {
			memcpy( &tmp_xy, &elmt_xy, sizeof(FXY_AREA) );
/*
			tmp_xy.x += THICKNESS/2;
			tmp_xy.y += THICKNESS/2;
			tmp_xy.w -= THICKNESS;
			tmp_xy.h -= THICKNESS;
*/
			draw_rect_mod( &tmp_xy, MOD_IN, rectangle->roundness, THICKNESS );
		} else {
			memcpy( &tmp_xy, &elmt_xy, sizeof(FXY_AREA) );
			tmp_xy.x += PH(2);
			tmp_xy.y += PV(2);
			tmp_xy.w -= PH(4);
			tmp_xy.h -= PV(4);
			draw_rect_mod( &tmp_xy, MOD_OUT, rectangle->roundness, PH(2) );
			draw_rect_mod( &elmt_xy, MOD_IN, rectangle->roundness, PH(2) );
		}
		element->color = TRANSPARENT;
		break;
	}
//	draw_options( element, &elmt_xy );
	draw_label( label, element, 0 );
}


void Draw_symbol( element )
	HDR *element;
{
	register SYMBOL_ELEMENT 	*symbol;
	FXY_AREA 					elmt_xy;	

	symbol = (SYMBOL_ELEMENT *)element_ptr( element );
	xy_map( (IRECT_AREA *)&element->row, &elmt_xy );

	if( element->fill && !(flags & STAT_SHADOWING) )
		draw_rect( &elmt_xy, 0, element->fill, 0, 0, 0 );

	draw_options( element, &elmt_xy );
	draw_symbol( element, symbol, &elmt_xy );
	draw_label( label, element, 0 );
}


void Draw_msgs( element )
	HDR *element;
{
	FXY_AREA elmt_xy;
	char *string;
#if 0
	IRECT_AREA area;
#endif
	
	if( element == NULL ) {
#if 0
		area.row = pict_size.height + PV(20);
		area.width = strlen( annotate_string )*font[0].map.width;
		if( pict_size.width > area.width )
			area.col = pict_size.width - area.width;
		else
			area.col = 0;
		area.height = PV(13);
		xy_map( &area, &elmt_xy );
		string = strformat( annotate_string );
		buf_printf("(%s) %d %d /QWFonty0 %s text\n",string, elmt_xy.x, elmt_xy.y, ps_colour( BLACK ) );
#endif
		elmt_xy.y = 16;
		elmt_xy.w = strlen( annotate_string )*8;
		if( paper_sizes[paper].width > elmt_xy.w )
			elmt_xy.x = paper_sizes[paper].width - elmt_xy.w - margin.w;
		else
			elmt_xy.x = 0;
		string = strformat( annotate_string );
		buf_printf("(%s) %d %d /QWFontX %s text\n",
			string, elmt_xy.x, elmt_xy.y, ps_colour( BLACK ) );
	}
}

void Draw_text( element )
	HDR *element;
{
	register TEXT_ELEMENT 	*text;
	FXY_AREA 				elmt_xy, bbox_xy;
	RECT_AREA 				bbox;
	int 					i;

	text = (TEXT_ELEMENT *)element_ptr( element );

#if 0
	if( !( element->attr & ENCLOSED ) && ( element->attr & (HIGHLIGHT|SELECTED ) ) ) {
		draw_shadow( element, WHITE );
	}
#endif

	xy_map( (IRECT_AREA *)&element->row, &elmt_xy );

	i = find_font( element, text,text->face?text->string+strlen(text->string)+1:NULL );


//..........Modif for enclose text by Guy Grimard at Multimeg............

	/* Determine text's bounding box (Add some extra room) */

	bbox.row = element->row - font[i].map.baseline - PV(2);
	bbox.col = element->col - PH(4);
	bbox.height = font[i].map.height; // + PV(2);

	/* ???? 13pt font is treated differently */
	if( i == 2 ) {
		bbox.row += PV(1);
		bbox.height -= PV(2);
	}
//.......................................................................

	/*  element->_width is used as a flag to say that we have
		already figured out this text's width
	*/
	if ( element->_width )
		bbox.width = element->width - PH(5); /* Don't ask why */
	else
		bbox.width = font[i].map.width * strlen(text->string) + PH(7);

	xy_map( (IRECT_AREA *)&bbox, &bbox_xy );

	if( element->fill != TRANSPARENT && !(flags & STAT_SHADOWING) )
		draw_rect( &bbox_xy, 0, element->fill, 0, 0, 0 );

//..........Modif for enclose text by Guy Grimard at Multimeg............

	bbox_xy.w-=PH(1);

//.......................................................................
	draw_options( element, &bbox_xy );

	if( !( element->attr & ENCLOSED ) && ( element->attr & (HIGHLIGHT|SELECTED ) ) ) 
		{
		bbox_xy.x -= PH(3);
		bbox_xy.y -= PV(3);
		bbox_xy.w += PH(6);
		bbox_xy.h += PV(6);
		draw_rect( &bbox_xy, BLACK, 0, PV(9), 0, 0 );
		}
	draw_text( text->string, &elmt_xy, element->color, element, text, i);
	draw_label( label, element, element->row );
}

#if 0
log_group( ptr )
	GROUP_ELEMENT *ptr;
{
	flog("list:%u\nscroll:%u\nvertical:%u\ntitle:%u\nlevel:%u\nexclusive:%u\n",
		ptr->list,ptr->scroll,ptr->vertical,ptr->title,ptr->level,ptr->exclusive);

	flog("zero:%u\ninternal:%u\nflexible:%u\nnon_standard:%u\nsame_width:%u\n",
		ptr->zero,ptr->internal,ptr->flexible,ptr->non_standard,ptr->same_width);

	flog("dont_change:%u\ninvert:%u\nspecial:%u\nno_attr:%u\nscroll_lock:%u\n",
		ptr->dont_change,ptr->invert,ptr->special,ptr->no_attr,ptr->scroll_lock);

	flog("type%u\nedge%u\nparm%u\n",
		ptr->type,ptr->edge,ptr->parm);
}
#endif
