/***********************************************************
    PROGRAM:    pictps
    MODULE:     image.c

    AUTHOR:     Patrick Hayden
    DATE:       1991-10-29

    DESCRIPTION:

	Render QNX Windows symbols, matrix images and scanmaps
	in PostScript.

************************************************************/

#include "pictps.h"
#include "bitmap.h"
#include "externs.h"

static bool 	mto;

void draw_symbol( element, symbol, xy )
	HDR 			*element;
	SYMBOL_ELEMENT 	*symbol;
	FXY_AREA 		*xy;
{
	int 			h,w;
	int     		r, c, i, n, nbytes, color;
	int    	 		*colors, nplanes = 0, alt_nplanes;
	char    		byte;
	register char 	*bitmaps;

	nplanes    = symbol->n;
	alt_nplanes = symbol->n_alt;
	colors     = (int *) symbol->colors;
	bitmaps    = (char *) (colors + symbol->n );
	h = symbol->h, w = symbol->w;

/*
	if (element->fill)
	    buf_printf( "%s %d %d %d %d frect\n", ps_colour(element->fill),
		   xy->w, xy->h, xy->x, xy->y );
*/

	nbytes = h * ((w + 7)/8);
	for( i = 0; i < nplanes; ++i ) {
	    color = colors[i];
	    if( color == TRANSPARENT ) color = element->color;

#if 0
	    buf_printf( "%d %d true [%d 0 0 %d 0 %d] {<\n",
	        w, h, w, -h, h);
#endif
	    buf_printf( "%d %d true [%d 0 0 %d 0 %d] {<\n",
	        w, h, w, h, 0);

		if( color != TRANSPARENT ) {
		    r = c = 0;
    		for( n = 0; n < nbytes; ++n ) {
    		    byte = bitmaps[i*nbytes + n];
        		buf_printf("%.2X",byte & 0xff);
		        if( (c += 8) >= w ) {
    	    	    c = 0;
        	    	++r;
	        	}
	    	}
		   	buf_printf( ">} %s %d %d %d %d IM\n", ps_colour(color),
        		PH(w), PV(h), xy->x, xy->y);
		}
	}
}


void draw_image( element, image )
	HDR 			*element;
	IMAGE_ELEMENT 	*image;
{
	IMAGE_DEF *def;

	if (image->descriptor) {
		def = (IMAGE_DEF *)image->data;
		switch (def->format & 0xff) {		
			case 'M':	draw_matrix( element, image, def );		break;
			case 'D':	draw_pixmap( element, image, def );		break;
			default:	draw_scanmap( element, image, def );
		}
	}
}

STATIC draw_matrix( element, image, def )
	HDR 			*element;
	IMAGE_ELEMENT 	*image;
	IMAGE_DEF 		*def;
{
	register char 		*data;
	int 				run_len, colour, tmp;
	int 				x,y;
	unsigned 			row, col;

	data = (char *)image->data + def->size;

	buf_printf("Gs %u SLW\n", V_TPP );
	
	mto = YES;
	run_len = 0;
	row = element->row;
	col = element->col;

	colour = matrix_index(0,data,def);

	for(y=0; y<def->height; y++ ) {
		for(x=0; x<def->width;x++ ) {
			tmp = matrix_index( y * def->width + x, data, def );		
			if( tmp != colour ) {
				fill_run( row, col, run_len, colour );
				col += PH(run_len);
				colour = tmp;
				run_len = 1;
			} else
				++run_len;
		}
		if (colour)
			fill_run( row, col, run_len, colour );
		run_len = 0;
		row += V_TPP;
		col = element->col;
		mto = YES;
	}
	buf_printf( "Gr\n" );
}

/* re-write of this function ALOW */
STATIC draw_pixmap( element, image, def )
	HDR 			*element;
	IMAGE_ELEMENT 	*image;
	IMAGE_DEF 		*def;
{
	register char 	*data;
	char 			*pdata;
	int 			run_len;
	unsigned 		row, col;

    int 			i,j,value, temp;
    int 			byte, mask;
    int 			bit1, bit2, bit3, bit4;
	unsigned 		tmp_colour = -1;

	pdata = (char *)image->data + sizeof(IMAGE_DEF);

	buf_printf( "Gs %u SLW\n", V_TPP );
	
	mto = YES;
	run_len = 0;
	row = element->row;
	col = element->col;

    if( def->type == 256 ) {
		for( j=0; j < def->height; j++ ) {
			data = &pdata[ j * 4 * def->bytes_per_line ];
        	for( i=0; i<def->width; i++ ) {
	            byte = i/8;
    	        mask = 128;
        	    mask >>= i&7;
	            bit1 = mask & data[byte];
				temp = def->bytes_per_line;
        	    bit2 = mask & data[byte+temp];
				temp += def->bytes_per_line;
    	        bit3= mask&data[byte+temp];
				temp += def->bytes_per_line;
            	bit4= mask&data[byte+temp];
	            value=((bit4?8:0)+(bit3?4:0)+(bit2?2:0)+(bit1?1:0));
    	        value++;
        	    if( tmp_colour != value ) {
            	    fill_run( row, col, run_len, tmp_colour );
        	        col += run_len;
	                tmp_colour = value;
    	            run_len = 1;
            	} else
	              	++run_len;

			}
            if( tmp_colour )
                fill_run( row, col, run_len, tmp_colour );
			mto = YES;
			run_len = 0;
			row += V_TPP;
			col = element->col;
			tmp_colour = -1;
		}
 	} else {
		for( j=0; j < def->height; j++ ) {
			data = &pdata[ j * def->bytes_per_line ];
        	for( i=0; i < def->width; i++ ) {
            	value = (int)data[i];
            	value++;

            	if( tmp_colour != value ) {
                	fill_run( row, col, run_len, tmp_colour );
     	   			col += run_len;
        	        tmp_colour = value;
                	run_len = 1;
            	} else
                	++run_len;

			}

            if( tmp_colour )
               	fill_run( row, col, run_len, tmp_colour );
			mto = YES;
			run_len = 0;
			row += V_TPP;
			col = element->col;
			tmp_colour = -1;
       	}
	}
	buf_printf( "Gr\n" );
}

STATIC draw_scanmap( element, image, def )
	HDR 			*element;
	IMAGE_ELEMENT 	*image;
	IMAGE_DEF 		*def;
{
	register opcode 	op;
	register unsigned 	arg;
	register unsigned 	*ptr;
	RECT_AREA 			pos;	
	unsigned 			colour;
	char 				*last, byte1, byte2;

	ptr = (unsigned *) ( (char *)image->data + def->size );
	pos.row = element->row;
	pos.col = element->col;
	mto = YES;
	colour = element->color;
	last = (char *) ( ptr + def->length );

	buf_printf( "Gs %u SLW\n", V_TPP );

	do {
		/* Upper bits are the opcode */

		op = (unsigned char )( (*ptr >> 4) & SCAN_MASK);

		if ( *ptr & 0x0080 ) {
			/* Short calls */
			arg = *ptr & SCAN_MASK;
			ptr = (unsigned *) ( (char *)ptr + 1 );
		} else {
			/* Long calls */
			byte1 =  (char) (*ptr & SCAN_MASK );
			byte2 = (char) (*ptr >> 8);
			arg = (byte1 << 8) | byte2;
			++ptr;
		}

		switch (op) {
			case SHORT_RUN:
			case SCAN_RUN:
				fill_run( pos.row, pos.col, arg, colour );			
				pos.col += PH(arg);
				break;
			case SHORT_ROW:
			case SCAN_ROW:
				pos.row += PV(arg);
				mto = YES;
				break;
			case SHORT_COL:
			case SCAN_COL:
				pos.col += PH(arg);
				mto = YES;
				break;
			case SHORT_COLOR:
			case SCAN_COLOR:
				colour = arg + 1;
				break;
			case SHORT_RUN_RETURN:
			case SCAN_RUN_RETURN:
				fill_run( pos.row, pos.col, arg, colour );
				pos.col = element->col;
				pos.row += V_TPP;
				mto = YES;
				break;
		}

	} while ( *ptr );
	buf_printf( "Gr\n" );
}

STATIC fill_run( row, col, run_len, colour )
	unsigned row, col;
	unsigned run_len, colour;
{
	int x, y;

	if ( !run_len ) return ( 0 );

	if ( mto )
		pxy_map( &x, &y, col, row );

	if( colour == TRANSPARENT ) {
		if( mto ) {
			buf_printf( "%d %d Mto\n", x + PH(run_len), y );			
			mto = NO;
		}
		else
			buf_printf( "%d SCNn\n", PH(run_len) );

		return( 1 );
	}
	
	if( mto ) {
		buf_printf( "%s %d %d %d SCNr\n",ps_colour(colour),
			PH( run_len ), x, y );
		mto = NO;
	}
	else 
		 buf_printf("%s %d SCNl\n", ps_colour(colour), PH(run_len) );
	
	return( 1 );
}

STATIC matrix_index( index, data, def )
	unsigned 	index;
	char 		*data;
	IMAGE_DEF 	*def;
{
	register char *p = data;

	switch( def->bits_per_color ) {
		case 4:
			p += index/2;
			if( index % 2 )
				return( *p & 0x0F);
			else
				return( ( *p >> 4 ) & 0x0F );
			break;
		case 8:
			p += index;
			return( *p );
			break;
		default:
			p += index + index;
        	return( *p );
			break;
	}
}
	
