/***********************************************************
    PROGRAM:    pictps
    MODULE:     io.c

    AUTHOR:     Patrick Hayden
    DATE:       1991-10-29

    DESCRIPTION:

	Speed up output by buffering writes.
************************************************************/
#ifdef __WATCOMC__
#include <stdarg.h>
#endif
#include "pictps.h"
#include "externs.h"

/* Buffered output */
static char 	cache[8000], *current = cache;
static char 	print_buf[1000];
static unsigned	size = 0;
static int		mode = 1;

#ifdef __WATCOMC__
/* Variable args under ANSI C */
static va_list	args;
#endif

#ifdef __WATCOMC__
void buf_io_ctrl( int io_mode )
#elif !defined __WATCOMC__
void buf_io_ctrl( io_mode )
	int io_mode;
#endif
{
	mode = io_mode;
}

char *buf_puts ( string )
	char 	*string;
{
	buf_strcat( string );
	return ( string );
}

#ifdef __WATCOMC__
int buf_printf( fmt, ... )
	char 	*fmt;
{
	va_start ( args, fmt );
	vsprintf( print_buf, fmt, args );
	va_end ( args );
	buf_strcat ( print_buf );
	return ( strlen (print_buf) );
}
#elif !defined __WATCOMC__
int buf_printf( fmt, arg )
	char 	*fmt;
	int 	arg;
{
	sprintf(print_buf, "%r", &fmt);
	buf_strcat( print_buf );
	return( strlen (print_buf) );
}
#endif

void buf_strcat ( src )
	char 	*src;
{
	unsigned len;

	if( mode == 0 )
		return;
	len = strlen( src );
	if ((size + len) >= 7000)
		flush_buffer();

	strcpy ( current, src );
	current += len;
	size += len;
}

void flush_buffer ( VOID )
{
#if 0
	fput(cache,size,stdout);
#endif
	fwrite( cache, size, 1, stdout );
	current = cache;
	cache[0] = '\0';
	size = 0;
}

unsigned short get_file_extension( VOID )
{
	char *offset;

	if ( ifilename ) {
	    if ((offset = strrchr(ifilename,'.'))==NULL)
    	    offset = ifilename + strlen (ifilename);

		/* Choose read format for extension - .pict, .wnd, .prt, other */
    	
		if ( offset != NULL ) {
			if ( !strcmp(offset,".pict") )
				return( STAT_PICT );
			if ( !strcmp(offset,".wnd") )
				return( STAT_WND );
			else if ( !strcmp(offset,".prt") )
				return( STAT_PRT );
		}
	}
	return( 0 );
}

determine_file_type( fp )
	FILE *fp;
{
	char str[10];

	if ( fread( str, strlen(PICTURE_HDR_ID), 1, fp ) ) {
		if ( !strncmp( PICTURE_HDR_ID, str, strlen(PICTURE_HDR_ID) ) ) 
			flags |= STAT_PICT;
		else
			flags |= STAT_PRT;
		fseek( fp, 0L, SEEK_SET );
		return( 1 );
	}
	return( 0 );
}


#ifdef __WATCOMC__
void error( fmt, ... )
	char 	*fmt;
{
	va_start ( args, fmt );
	vfprintf(stderr,fmt,args);
	va_end ( args );
	exit( -1 );
}
#elif !defined __WATCOMC__
void error( fmt, arg )
	char 	*fmt;
	int		arg; 
{
	fprintf( stderr, "%r", &fmt );
	exit( -1 );
}
#endif

#ifndef __WATCOMC__
int isatty( fd )
	unsigned fd;
{
	return( ftty( stdin ) );
}
#endif 

#ifdef __WATCOMC__
int flog ( fmt, ... )
	char *fmt;
{
	if( !( options & OPT_VERBOSE ) )
		return( 0 );

	va_start ( args, fmt );
	if ( options & OPT_VERBOSE ) 
		vfprintf( stderr, fmt, args );
	va_end ( args );

	return ( 1 );
}
#elif !defined __WATCOMC__
int flog ( fmt, arg )
	char	*fmt;
	int		arg;
{
	if ( options & OPT_VERBOSE )
		return ( fprintf( stderr, "%r", &fmt ) );
	return ( 0 );
}
#endif
