/***********************************************************
	PROGRAM:	pictps
	MODULE:		main.c

	AUTHOR:		Patrick Hayden
	DATE:		1991-10-29

	DESCRIPTION:

	Reads QNX Windows picture files and prt files and
	tries to make sense of them.
************************************************************/

#include "pictps.h"
#include "externs.h"

#ifdef __USAGE
%C [options] file > output
%C [options]

Options:
-a msg				Annotate the printout with a message 
-b bkgd				Force the background colour
-c filename			RGB mapping file
-e					Output in Encapsulated PostScript format (default)
-H					Attempt to print in Helvetica
-l filename			File to search for link pictures
-m					Write minimally conforming EPS
-n nid				Attempt to connect to QNX Windows by nid
-o					One to one mapping
-p papersize		0: Letter 	1: Landscape letter 
					2: Legal	3: Landscape Legal	
-P					Output straight PostScript format, not eps
-q					Query Windows for palette, font and scaling information
-s					Black window outline
-v					Verbose mode (output on stderr)
#endif

bool				pinned = NO;

static PRINTHDR		printhdr;

static FILE			*fp;
static char			hdr_buf[PICTURE_HDR_SIZE];
static PICTURE_HDR	*hdr = hdr_buf;

/* These should always point to the same buffer */
static PICTHDR		*picthdr, *windhdr;
static HDR			*element;
static unsigned		element_buff_size;

static long			new_wind_offset = 0;
static RECT_AREA	temp_area;

#ifndef __WATCOMC__
extern char *strchr();
#endif

main( argc, argv )
	int argc;
	char *argv[];
{
#ifdef __WATCOMC__
	if( need_usage( argv ) ) print_usage( argv );
#elif !defined __WATCOMC__
	if ( argc < 2 )	show_usage();
#endif
        
	SetName("pictps",NULL);

	init_vars();
	extract_options( argc, argv );

	if( !ifilename ) {
		if( isatty( 0 ) )
			error( "Cannot read stdin from a tty\n");
		ifilename = make_temp_file();
		flog( "Created tmp file: %s\n", ifilename );
	}

	if ((fp = fopen(ifilename,"r"))==NULL)
		error("Could not open file: %s\n",ifilename);

	element = (HDR *)salloc( ELEMENT_BUFF_SIZE );
	element_buff_size = ELEMENT_BUFF_SIZE;

	/* Alias the buffer */
	picthdr = windhdr = (PICTHDR *)element;

	Init_coords();

	flags |= get_file_extension();
	if( !( flags & (STAT_PICT|STAT_WND|STAT_PRT) ) )
		determine_file_type( fp );

	if ( flags & STAT_PRT ) {
		if ( ( options & OPT_DISPLAY ) && !( flags & STAT_GRAPHICS ) )
			open_graphics( 0 );
		read_PRINTHDR( fp );
	} else
		read_PICTURE_HDR( fp );


	if( options & OPT_QUERY_SCREEN ) {
		query_font_info();
		query_map_rgb();
	}

	if( options & OPT_SET_PALETTE )
		read_colours( palette_file );

	switch ( flags & ( STAT_PICT|STAT_WND|STAT_PRT ) ) {
		case STAT_PICT:
			ps_write_hdr();
			ps_draw_bkgd(hdr);
			read_picture();
			break;
		case STAT_WND:
			ps_write_hdr();
			ps_draw_bkgd(hdr);
			read_window();
			break;
		case STAT_PRT:
			read_prt_file();
			break;
		default:
			error( "File type undefined\n" );
	}

	fclose( fp );

	if( flags & STAT_TMPFILE )
		unlink( ifilename );

	ps_write_trailer();
	flush_buffer();
	
	/* XXXXX */
	if( options & OPT_DISPLAY )
		sleep( 10 );

	flog( "Exiting\n" );
	exit( 0 );
}

STATIC char *make_temp_file( VOID )
{
	FILE *outfp;
	char *buf;

	buf = (char *)malloc( 30 );
#ifdef __WATCOMC__
	sprintf( buf, "/tmp/pictps%d", getpid() );
#elif !defined __WATCOMC__
	sprintf( buf, "/tmp/pictps%d", My_pid );
#endif
	if( ( outfp = fopen( buf, "w" ) ) == NULL )
		error("Could not open tmp file for write: %s\n", buf );
	do {
		fread( path, 280, 1, stdin );
		fwrite( path, 280, 1, outfp );
	} while( !feof( stdin ) );
	fclose( outfp );
	flags |= STAT_TMPFILE;
	return( buf );
}

STATIC void read_PICTURE_HDR( fp )
	FILE *fp;
{
	fget ( hdr, PICTURE_HDR_SIZE, fp );

	if ((strcmp(hdr->id,PICTURE_HDR_ID))||(!hdr->picture))
		error("Not a picture file\n");

	/* Extract height and width from header */
	if ( hdr->height && hdr->width ) {
		pict_id = 0;
		pict_size.height = hdr->height;
		pict_size.width = hdr->width;
	} else {
        if( options & OPT_QUERY_SCREEN ) 
			query_pict_size( ifilename );    	
		else {
			pict_size.height = PV(10);
			pict_size.width = PH(10);
			while( read_HDR( fp ) ) {
				pict_size.height = 
					max( pict_size.height, element->row + element->height );
				pict_size.width =
					max( pict_size.width, element->col + element->width );
				fseek( fp, (long)( element->length - sizeof(HDR) ), 1 );
			}
			fseek( fp, (long)PICTURE_HDR_SIZE, 0 );
		}
	} 
				
}

STATIC void read_PRINTHDR( fp )
	FILE *fp;
{
	int i;

	if ( !fget ( &printhdr, sizeof(PRINTHDR), fp ) )
		error( "Could not extract print header\n" );

	if ( printhdr.record_type != FILE_HEADER )
		error( "%s: not an extended window file\n", ifilename );

	pict_size.height = printhdr.height;
	pict_size.width = printhdr.width;

	for (i=1;i<=printhdr.nbr_colors;i++) {
		rgb_map[i].r = printhdr.color[i-1].red   / 10000.0;
		rgb_map[i].g = printhdr.color[i-1].green / 10000.0;
		rgb_map[i].b = printhdr.color[i-1].blue  / 10000.0;
	}
	
	/* Nice whites */
	if ( (rgb_map[5].r + rgb_map[5].g + rgb_map[5].b) > ( 3 * 0.9 ) )
		rgb_map[5].r = rgb_map[5].g = rgb_map[5].b = 1.0;

}


STATIC read_HDR( fp )
	FILE *fp;
{
	return( fget( element, sizeof(HDR), fp ) );
}

STATIC read_element( fp )
	FILE *fp;
{
	int *iptr;
	unsigned short len;


	/* Read picture header and return 0 on EOF */
	if( ( len = read_HDR( fp ) ) < sizeof(HDR) )
		return( 0 );

	if( !element->length )
		return( 0 );

	if( element->length > element_buff_size ) {
		flog( "Growing element buffer: %u bytes ->%u bytes\n", 
			element_buff_size, element->length );

		iptr = (int *)malloc( element->length );
		if( iptr == NULL )
			error( "malloc failed while growing element buffer\n" );
		memmove( iptr, element, sizeof( HDR ) );
		free( element );
		element = (HDR *)iptr;
		windhdr = picthdr = (PICTHDR *)iptr;
	}
	
	/* One of the print structures */
	if ( (element->length < len) ) {
		fseek( fp, -(long)(len - element->length), 1 );
		return( 1 );
	}
		
	/* Read extra element info, if any */
	if (element->length - sizeof(HDR))
		if ( fget((char *)element+len,element->length-sizeof(HDR),fp) != 
			element->length - sizeof( HDR ) )
				return(0);

	/* Set this in case we are displaying the element */
	iptr = (int *) ( (char *) element + element->length );
	*iptr = 0;
	
	return( 1 );
}

STATIC int read_colours( fname )
	char *fname;
{
	FILE *fp;
	char buf[80], *p, *p1;
	int i;
	struct rgb trgb;

	if ((fp = fopen (fname,"r")) == NULL)
		error("Could not open RGB file: %s\n",fname);

	while ( fgetline (fp, buf, 80) ) {
		if( isalpha( buf[0] ) ) {
			/* .cfg file - ignore color names */
			StrTok( buf );
			trgb.r = atof( StrTok( NULL ) ) / 255.0;
			trgb.g = atof( StrTok( NULL ) ) / 255.0;
			trgb.b = atof( StrTok( NULL ) ) / 255.0;
			i = atoi( strtok( NULL,"	 " ) );
			if( i > 0 && i < 257 )
				memcpy( &rgb_map[i], &trgb, sizeof( struct rgb ) );
		} else {
			p = strchr(buf,0x20);
			*p = '\0';
	
			i = atoi(buf);
   
			/* Repeatedly search for a space -> NULL -> atof */

			if ((p1 = strpbrk(p + 1, " \n")) == NULL)
					error ("Syntax error in colour map file: %s\n",fname);

			*p1 = '\0';
			rgb_map[i].r = atof(p + 1);
			p = p1;

			if (p1 = strchr(p + 1, 0x20)) {
				*p1 = '\0';
				rgb_map[i].g = atof(p + 1);
				p = p1;
			} else {
				gray_level(i);
				continue;
			}

			if (p1 = strpbrk(p + 1, " \n")) {
				*p1 = '\0';
				rgb_map[i].b = atof(p + 1);
			} else {
				gray_level(i);
				continue;
			}
		}
	}

	fclose(fp);
	return( 1 );           
}

STATIC gray_level( i )
	int i;
{
	rgb_map[i].g = rgb_map[i].b = rgb_map[i].r;
}

STATIC read_picture( VOID )
/*
 *	Called for .pict, .wnd and .prt files
 */
{
	while ( read_element( fp ) ) {
		flog("Type: %c,%d Tag: %s Length: %u\n",element->type,element->type,
			(char *)element + sizeof(HDR), element->length );
		
		if ( (element->type == WIND_HEADER) || (element->type == PICT_HEADER) ||
			 (element->type == FILE_TRAILER) )
	    	return;

		Draw_element(element);
	}
}

STATIC read_window( VOID )
/*
 *	Called for both .wnd and .prt files 
 */
{
	TEXT_ELEMENT 				*text;
	static char 				windopts[50], *paneopts[4];
	long 						prev;
	RECT_AREA 					frames[5];
	int 						i;

	for (i=0;i<=3;i++) {
		paneopts[i] = salloc(50);
		paneopts[i][0] = '\0';
		wind.pane[i].picture = 0;
	}

	if ( flags & STAT_SCREEN ) {
		temp_area.row = windhdr->row;
		temp_area.col = windhdr->col;
		add_offset(&temp_area);
	}

	prev = ftell(fp);

	while( read_element( fp ) ) {
		flog("Scanning: Type: %c,%d Tag: %s\n",element->type,element->type,
			(char *)element + sizeof(HDR));

		if ( (element->type == WIND_HEADER) || (element->type == PICT_HEADER) ||
		 (element->type == FILE_TRAILER) ) {
			new_wind_offset = prev;
	    	break;
		}

		switch(CASE(TAG(element)+1)) {
		case 'TO':
			wind.bar.top = prev;
			frames[BTOP].height = element->height;
			skip_to_group_end();
			break;

		case 'BO':
			wind.bar.bottom = prev;
			frames[BBOTTOM].height = element->height;
			skip_to_group_end();
			break;
			
		case 'RI':
			wind.bar.right = prev;
			frames[BRIGHT].width = element->width;
			skip_to_group_end();
			break;

		case 'LE':
			wind.bar.left = prev;
			frames[BLEFT].width = element->width;
			skip_to_group_end();
			break;

		case 'PI':
			*(TAG(element) + 9) = '\0';
			i = atoi( TAG( element ) + 8 );
			wind.pane[i].picture = prev;
			memcpy( &wind.pane[i].view, &element->row, sizeof( RECT_AREA ) );
			skip_to_group_end();
			break;

		case 'IC':
			wind.icon = prev;
			skip_to_group_end();
			break;

		case 'SI':
			memmove (&wind.size,&element->height,sizeof(BOX));
			break;

		case 'OP':
			text = (TEXT_ELEMENT *)element_ptr( element );
			flog("Extracting window options: %s\n",text->string);
			strcpy(windopts,text->string);
			break;
			
		case 'PA':
			if (element->type == RECTANGLE) {
				*(TAG(element) + 7) = '\0';
				memmove(&wind.pane[atoi(TAG(element) + 6)].pos,&element->row,sizeof(RECT_AREA));
			} else {
				text = (TEXT_ELEMENT *)element_ptr(element);
				*(TAG(element) + 6) = '\0';
				flog("Extracting pane options: %s\n",text->string);
				strcpy(paneopts[atoi(TAG(element) + 5)],text->string);
			}
			break;
		
		case 'TI':
			text = (TEXT_ELEMENT *)element_ptr(element);
			strcpy(wind.title,text->string);
			break;

		case 'ST':
			text = (TEXT_ELEMENT *)element_ptr(element);
			wind.state.active = wind.state.iconified = wind.state.pinned = NO;
			for (i=0;i<strlen(text->string);i++) {
				switch(text->string[i] & 0xff) {
					case 'A':	wind.state.active = YES;		break;		
					case 'I':	wind.state.iconified = YES;		break;
					case 'P': 	wind.state.pinned = YES;
								pinned = YES;					break;
				}
			}
			break;

		case 'CO':
			break;

		case 'HI':
			break;

		default:
			if (element->type == GROUP)
				skip_to_group_end();
			break;
		}
		prev = ftell(fp);
	}


	frames[BTOP].width = frames[BLEFT].width + wind.size.width + frames[BRIGHT].width;
	frames[BTOP].height += PV(7);
	frames[BBOTTOM].width = frames[BTOP].width;
	frames[BLEFT].height = wind.size.height;
	frames[BRIGHT].height = frames[BLEFT].height;
	frames[BPANE].height = wind.size.height;
	frames[BPANE].width = wind.size.width;


	frames[BTOP].col = 0;
	frames[BTOP].row = 0;
	
	frames[BBOTTOM].col = frames[BTOP].col;
	frames[BBOTTOM].row = frames[BTOP].row + frames[BTOP].height + wind.size.height - 80;

	frames[BLEFT].col = frames[BTOP].col;
	frames[BLEFT].row = frames[BTOP].row + frames[BTOP].height - PV(7);

	frames[BRIGHT].col = frames[BTOP].col + frames[BLEFT].width + wind.size.width;
	frames[BRIGHT].row = frames[BLEFT].row;

	frames[BPANE].col = frames[BTOP].col + frames[BLEFT].width;
	frames[BPANE].row = frames[BTOP].row + frames[BTOP].height;

	wind.size.height += frames[BTOP].height + frames[BBOTTOM].height - PV(7);
	wind.size.width += frames[BLEFT].width + frames[BRIGHT].width;

	get_window_options(&wind, &windopts[0], &paneopts);

	if (wind.state.iconified)
		draw_icon(&wind);
	else
		draw_window(&wind, frames);

	for( i=0; i<=3; i++ )
		free( paneopts[i] );
}

STATIC draw_icon( wind )
	struct wind_struct *wind;
{
	FXY_AREA xy;
	USER_DATA *options_data;

	if ( flags & STAT_SCREEN )
		sub_offset(&temp_area);

	fseek( fp, wind->icon, 0 );
	read_element( fp );
	data_ptr( element );
	
	push_offset();
	add_offset( (RECT_AREA *)&element->row );
	element->row = element->col = 0;

	xy_map( (IRECT_AREA *)&element->row,&xy);

	if (element->fill == TRANSPARENT)
		element->fill = GRAY;

	if (element->fill != NO_COLOR)
		draw_rect(&xy,element->fill,element->fill,0,1,0);

	options_data = (USER_DATA *) ( (char *)data + data->data );
	if( strchr( (char *)options_data->data, 'b' ) || element->attr & MODELLED ) {
		draw_rect_mod( &xy, MOD_OUT, 0, H_TPP );

#if 0
		element->row = element->col = PH(2);
		add_offset( (RECT_AREA *)&element->row );
#endif
	}

	if( label->text[0] != '\0' ) {
		pxy_map( &xy.x, &xy.y,
			( element->width - strlen( label->text ) * font[2].map.width ) / 2,
			element->height - PV(8) );
		buf_printf( "(%s) %d %d /QWFont2 CC1 text\n",
			strformat( label->text ), xy.x, xy.y );
	
	}

	draw_to_group_end();

	pop_offset();
}

STATIC draw_window( wind, frames )
	struct wind_struct *wind;
	RECT_AREA frames[5];
{
	IRECT_AREA box;
	FXY_AREA xy, xy_tmp;
	int i;

	box.row = frames[BTOP].row;
	box.col = frames[BTOP].col;
	box.height = wind->size.height;
	box.width = wind->size.width;
	if( !( flags & STAT_SCREEN ) && !( options & OPT_BLACK_LINE ) ) {
        box.height += PH(6);
		box.width += PH(6);
	}
	xy_map( (IRECT_AREA *)&box,&xy);
	draw_rect( &xy, TRANSPARENT, LIGHT_GRAY, 0, 1, 0 );

	/* For modelled windows, shift here */
	if( !( flags & STAT_SCREEN ) && !( options & OPT_BLACK_LINE ) ) {
		box.row = PH(3);
		box.col = PV(3);
		add_offset( (RECT_AREA *)&box );
	}
	
	/* Set flag to indicate that a frame is being printed */
	flags |= STAT_FRAME;
	
	/* Top frame bar */
	fseek( fp, wind->bar.top, 0 );
	read_element( fp );
	memmove( &element->row, &frames[BTOP].row, sizeof(RECT_AREA) );
	draw_to_group_end();
	
	/* Bottom frame bar */
	fseek( fp, wind->bar.bottom, 0 );
	read_element( fp );
	memmove( &element->row, &frames[BBOTTOM].row, sizeof(RECT_AREA) );
	draw_to_group_end();

	/* Right frame bar */ 
	fseek( fp, wind->bar.right, 0 );
	read_element( fp );
	memmove( &element->row, &frames[BRIGHT].row, sizeof(RECT_AREA) );
	draw_to_group_end();

	/* Left frame bar */
	fseek( fp, wind->bar.left, 0 );
	read_element( fp );
	memmove( &element->row, &frames[BLEFT].row, sizeof(RECT_AREA) );
	draw_to_group_end();

	/* Back to normal printing */
	flags &= ~STAT_FRAME;

	for( i=0; i<=3; i++ )
		if( wind->pane[i].picture ) {
			xy_map( (IRECT_AREA *)&wind->pane[i].pos, &xy );
			if(wind->pane[i].opts & PANE_VSCROLL) {
				memcpy(&xy_tmp,&xy,sizeof(FXY_AREA));
				xy_tmp.x += xy_tmp.w + PH(6);
				draw_vscroll( &xy_tmp );
			}
			if(wind->pane[i].opts & PANE_HSCROLL) {
				memcpy( &xy_tmp, &xy, sizeof(FXY_AREA) );
				xy_tmp.y += xy_tmp.h + PH(6);
				draw_hscroll( &xy_tmp );
			}
			
			fseek( fp, wind->pane[i].picture, 0 );
			read_element( fp );

			/* Print picture contents */
			read_element( fp );

			if( options & OPT_DISPLAY ) {
				open_display_window( xy.h, xy. w);
				flags |= STAT_DISPLAYING;
			}

			xy_map( (IRECT_AREA *)&wind->pane[i].pos, &xy );
			draw_rect( &xy, 0, picthdr->phdr.bkgd, 0, 1, 0 );

#if 0
			if( flags & STAT_PRT ) {
				element->row = element->col = 0;
				element->height = wind->pane[i].pos.height;
				element->width = wind->pane[i].pos.width;
			} else {
				element->row = frames[BPANE].row;
				element->col = frames[BPANE].col;
			}
#endif

			push_offset();
			add_offset( &wind->pane[i].pos );
			if( flags & STAT_PRT ) {
				sub_offset( &wind->pane[i].view );
				element->row = wind->pane[i].view.row;
				element->col = wind->pane[i].view.col;
				element->height = wind->pane[i].pos.height;
				element->width = wind->pane[i].pos.width;
			} else {
				element->row = frames[BPANE].row;
				element->col = frames[BPANE].col;
			}
			draw_to_group_end();
			pop_offset();
#if 0
			sub_offset( &wind->pane[i].pos );
#endif
	
			flags &= ~STAT_DISPLAYING;

			/* Draw pane outline */
			if(wind->pane[i].opts & PANE_OUTLINE) {
				wind->pane[i].pos.row -= PV(2);
				wind->pane[i].pos.col -= PH(2);
				wind->pane[i].pos.height += PV(4);
				wind->pane[i].pos.width += PH(4);
				xy_map( (IRECT_AREA *)&wind->pane[i].pos, &xy );
				draw_rect_mod(&xy,MOD_IN,0,1);
			}

		}

	if (wind->windopts & OUTLINE_FRAME) {
		if (wind->windopts & MODEL_FRAME) {
			box.row = frames[BTOP].row;
			box.col = frames[BTOP].col;
			box.height = wind->size.height;
			box.width = wind->size.width;
			xy_map( (IRECT_AREA *)&box,&xy);
			if( options & OPT_BLACK_LINE )
				draw_rect( &xy, BLACK, 0, 0, 1, 0 );
			else
				draw_rect_mod( &xy, MOD_OUT, 0, 1 );
		} else {
			draw_rect( &xy, BLACK, 0, 0, 1, 0 );
		}
	}

	if( wind->windopts & RESIZE_TABS ) {
		box.row = frames[BTOP].row;
		box.col = frames[BTOP].col;
		xy_map( (IRECT_AREA *)&box,&xy);		
		buf_printf("%d %d cornerUL\n",xy.x, xy.y );
		xy.x += frames[BTOP].width - CORNER_W;
		buf_printf("%d %d cornerUR\n",xy.x,xy.y);

		box.row = frames[BBOTTOM].row;
		box.col = frames[BBOTTOM].col;
		xy_map( (IRECT_AREA *)&box, &xy );
		xy.y += frames[BBOTTOM].height - CORNER_H;
		buf_printf("%d %d cornerLL\n",xy.x,xy.y);
		xy.x += frames[BTOP].width - CORNER_W;
		buf_printf("%d %d cornerLR\n",xy.x,xy.y);
	}

	if( flags & STAT_SCREEN )
		sub_offset( &temp_area );
}

STATIC read_prt_file( VOID )
{
	bool screen_bkgd = NO;
	IRECT_AREA area;
	FXY_AREA xy;

	while( screen_bkgd?YES:read_element( fp ) ) {

		/* Screen bkgd picture is handled differently */
		screen_bkgd = NO;

		switch (element->type & 0xff) {
		case PICT_HEADER:
			/* scale used to indicate PrintArea landscape mode - ALOW */
			if( picthdr->scale == -1 && ( paper % 2 ) )  
				++paper;
			if( !( flags & STAT_SCREEN ) ) {
				flags |= STAT_PICTURE;
				pict_size.height = picthdr->height;
				pict_size.width = picthdr->width;
				ps_write_hdr();
			}
			if( options & OPT_DISPLAY ) {
				open_display_window( pict_size.height, pict_size.width );
				flags |= STAT_DISPLAYING;
			}
			ps_draw_bkgd( &picthdr->phdr );
			read_picture();
			flags &= ~STAT_PICTURE;
			break;
            
		case WIND_HEADER:
			if( ! screen_bkgd ) flags |= STAT_WINDOW;
			if( !( flags & STAT_SCREEN ) ) {
				pict_size.height = windhdr->height;
				pict_size.width = windhdr->width;
				ps_write_hdr( );
			}
			ps_draw_bkgd( &windhdr->phdr );
			read_window();
			flags &= ~STAT_WINDOW;
			break;

		case SCRN_HEADER:
			flags |= STAT_SCREEN;
			if( picthdr->height && picthdr->width ) {
				pict_size.height = picthdr->height;
				pict_size.width = picthdr->width;
			} else {
				pict_size.height = PH(768);
				pict_size.width =  PV(1024);
			}
			ps_write_hdr();
			if( !windhdr->phdr.bkgd ) windhdr->phdr.bkgd = DARK_GRAY;
			ps_draw_bkgd( &windhdr->phdr );
			area.row = area.col = 0;
			area.height = pict_size.height;
			area.width = pict_size.width;
			xy_map( &area, &xy );
			boxclip_start( &xy );
			if( read_element( fp ) ) {
				read_picture();
				screen_bkgd = YES;
			}
			break;

		case FILE_TRAILER:
			if( flags & STAT_SCREEN ) clip_end();
			break;

		default:
				error("\nFile corrupt... Type: %c Tag: %s ",element->type,
					(char *)element + sizeof(HDR) );
		}

		if( new_wind_offset ) {
			fseek( fp, new_wind_offset, 0 );
			new_wind_offset = 0;
		} else
			if( !screen_bkgd )
				break;

	}

}

STATIC skip_to_group_end( VOID )
{
	int level = 1;

	while( level && read_HDR( fp ) ) {
		fseek( fp, (long)( element->length - sizeof(HDR ) ), 1 );
		if( element->type == GROUP )
			++level;
		else if( element->type == END_OF_GROUP )
			--level;
	}
}

STATIC draw_to_group_end( VOID )
{
	FXY_AREA xy;
	int level = 1;

	xy_map( (IRECT_AREA *)&element->row,&xy);
	boxclip_start ( &xy );
	Draw_element( element );
	while( level && read_element( fp ) ) {
		flog("Drawing: Type: %c,%d Tag: %s\n",element->type,element->type,
			(char *)element + sizeof(HDR));
    	if( element->type == GROUP )
			++level;
		else if( element->type == END_OF_GROUP )
			--level;
    	Draw_element( element );
	}
	clip_end();
}

STATIC open_display_window( height, width )
	unsigned height, width;
{
	static unsigned display_wid = 0;
	static unsigned count = 0;
	
#if 0
	if( display_pid ) {
		WindowCurrent( display_wid );
		WindowClose();
		PictureClose( display_pid );
	}
#endif

	sprintf( tbuf, "Pictps%d", ++count );
    display_pid = PictureOpen( tbuf, NULL, tbuf, GRAY, 0, 
		NULL, NULL );
	display_wid = WindowOpen( tbuf, height, width, NULL, 
		NULL, NULL, 0 );
}

STATIC void init_vars( VOID )
{
	H_TPP = V_TPP = 11;

	font[0].sys.height = 132;
	font[0].sys.width = 88;
	font[0].sys.baseline = 99;
	font[0].sys.ps_font = "Courier-Bold";
	font[0].sys.ps_sx = 141;
	font[0].sys.ps_sy = -175;
	font[1].sys.height = 165;
	font[1].sys.width = 88;
	font[1].sys.baseline = 121;
	font[1].sys.ps_font = "Courier-Bold";
	font[1].sys.ps_sx = 141;
	font[1].sys.ps_sy = -175;
	font[2].sys.height = 143;
	font[2].sys.width = 66;
	font[2].sys.baseline = 110;
	font[2].sys.ps_font = "Courier";
	font[2].sys.ps_sx = 105;
	font[2].sys.ps_sy = -140;
	num_fonts = 3;
}

#ifdef __WATCOMC__
STATIC extract_options( argc, argv )
	unsigned argc;
	char *argv[];
{
	int j, c;
	int errflag = 0;

	/* This is now the default */
	options |= OPT_EPS;

    while( ( c = getopt( argc, argv, "a:b:c:deD:Hl:mn:op:Pqsv" ) ) != -1 ) {
		switch( c ) {
			case 'a':	annotate_string = optarg;		
						margin.y += 36;
						options |= OPT_ANNOTATE;		break;
			case 'b':	bcol = atoi( optarg );
						options |= OPT_FORCE_BKGD;		break;
			case 'c':	palette_file = optarg;
						options |= OPT_SET_PALETTE;		break;
			case 'd':	options |= OPT_DISPLAY;			break;
			case 'D':	drsikes = atoi( optarg );		break;
			case 'e':	options |= OPT_EPS;				break;
			case 'H':	options |= OPT_HELVETICA;
						for( j=0 ; j<=2; j++ ) {
							font[j].sys.ps_font = "Helvetica";
							font[j].sys.ps_sx *= 1.2;
						}								break;														break;
			case 'l':	link_open( optarg );			break;
			case 'm':	options |= OPT_EPS|OPT_MIN_EPS;	break;
			case 'n':	open_graphics(atoi( optarg ));	break;
			case 'o':	options |= OPT_ONE_TO_ONE;		break;
			case 'p':	paper = atoi( optarg );			break;
			case 'P':	options &= ~(OPT_EPS|OPT_MIN_EPS);
														break;
			case 'q':	options |= OPT_QUERY_SCREEN;	break;
			case 's':	options |= OPT_BLACK_LINE;		break;
			case 'v':	options |= OPT_VERBOSE;			break;
			case '?': 	++errflag;						break;
		}
	}

	if ( optind == argc )
		ifilename = NULL;
	else
		ifilename = argv[optind];

	if( errflag ) print_usage( argv );

#if 0
	/* Do not work with QNX 4 static lib */
	if( !( flags & STAT_GRAPHICS ) )
		 H_TPP = V_TPP = 11;
#endif
	return( YES );
}
#elif !defined __WATCOMC__
STATIC extract_options( argc, argv )
	unsigned argc;
	char *argv[];
{
	int i, j;

	ifilename = argv[1];

	/* defaults */
	options |= OPT_EPS;

	for( i=2; i < argc; i++ ) {
		switch( CASE( argv[i] ) ) {
    	case 'ml':	margin.x = atoi(&argv[i][3]);	break;
		case 'mr':	margin.w = atoi(&argv[i][3]);	break;                                                                                                                                                                                                                                                                                                                                                                                                                                                                     
		case 'mt':	margin.h = atoi(&argv[i][3]);	break;                                                                                                                                                                                                                                                                                                                                                                                                                                                                     
		case 'mb':	margin.y = atoi(&argv[i][3]);	break;                                                                                                                                                                                                                                                                                                                                                                                                                                                                     
		case 'p=':	paper = atoi(&argv[i][2]);		break;
		case 'a=':	annotate_string = &argv[i][2];	
					margin.y += 36;
					options |= OPT_ANNOTATE;		break;
		case 'b=':	bcol = atoi(&argv[i][2]);
					options |= OPT_FORCE_BKGD;		break;
		case 'D=':	drsikes = atoi( &argv[i][2]);	break;
		case 'n=':	open_graphics(atoi(&argv[i][2]));
													break;
		case 'l=':	link_open(&argv[i][2]);			break;
		case 'c=':	palette_file = &argv[i][2];
					options |= OPT_SET_PALETTE;		break;
		case '+d':	options |= OPT_DISPLAY;			break;
		case '+1': 	options |= OPT_ONE_TO_ONE;		break;
		case '+e':	options |= OPT_EPS;				break;
		case '+m':	options |= OPT_EPS|OPT_MIN_EPS;	break;
		case '+H':	options |= OPT_HELVETICA;
					for( j=0 ; j<=2; j++ ) {
						font[j].sys.ps_font = "Helvetica";
						font[j].sys.ps_sx *= 1.2;
					}								break;
		case '+P':	options &= ~(OPT_EPS|OPT_MIN_EPS);
													break;
		case '+q':	options |= OPT_QUERY_SCREEN;	break;
		case '+s':	options |= OPT_BLACK_LINE;		break;
		case '+v':	options |= OPT_VERBOSE;			break;

		default: 	error ("Invalid option\n"); 	break;
		}
	}

#if 0
	/* 	This does not appear to work with QNX static lib
		...they're not changing anyways */
	if ( !( flags & STAT_GRAPHICS ) )
     	H_TPP = V_TPP = 11;
#endif

	if( options & OPT_ANNOTATE ) {
		if( paper % 2 ) 
			margin.w += 36;
		else
			margin.y += 36;
	}
	return (YES);
}
#endif

#ifndef __WATCOMC__
STATIC show_usage( VOID )
{
	int i;
	printf("Use: pictps inputfile [options] > outputfile\n\n");
	printf("Options: p=papersize  Paper sizes in 1/72\":\n");
	for (i=0;i<4;i++)
	printf("						%d: %s %ux%u\n",i,paper_sizes[i].name,
			paper_sizes[i].width,paper_sizes[i].height);
	printf("		 ml=left margin mr=right margin mt=top margin mb=bottom margin\n");
	printf("		 a=msg b=bkdg colour n=screen_node +display_prt\n");
	printf("		 c=RGB file l=link file +1to1mapping +eps(nop) +min_eps\n");
	printf("		+Helvetica +simple_outline +query_screen +verbose\n");
	printf("		+PostScript(not EPSF)\n");
	exit(0);
}
#endif

