/************************************************************
    PROGRAM:    pictps
    MODULE:     misc.c

    AUTHOR:     Patrick Hayden
    DATE:       1991-10-29

    DESCRIPTION:

	Miscellaneous functions from here and there
************************************************************/

#include "pictps.h"
#include "externs.h"

void draw_shadow( element, color )
	HDR *element;
	unsigned color;
{
	HDR *eptr;

	/* do not allow infinite recursion */
	if( flags & STAT_SHADOWING )
		return;

	/* Copy element, change attributes so shadowed, draw */
	eptr = (HDR *)salloc( element->length );
	memmove( eptr, element, element->length );
	eptr->color = eptr->fill = color;
	eptr->row += PV(2);
	eptr->col += PH(1);
	eptr->attr &= ~SHADOWED;
	eptr->flags &=  ~EXTRA_DATA;
	data = data_ptr( eptr );

	/* Set flag in case element specific */
	flags |= STAT_SHADOWING;
	Draw_element( eptr );
	free( eptr );
	flags &= ~STAT_SHADOWING;
	data = data_ptr( element );
}


void draw_options( element, bbox )
	HDR 		*element;
	FXY_AREA 	*bbox;
{
	if( element->attr & ENCLOSED ) 
		{
		bbox->x -= PH(3);
		bbox->y -= PV(3);
		bbox->w += PH(6);
//		bbox->h += PV(6);
		bbox->h += PV(6);

		if ( element->attr & MODELLED ) 
			{
			/* Decided whether depressed */
			if((element->attr&(SELECTED|HIGHLIGHT))||(element->flags&STATE_SET))
				{
				if( element->type == TEXT ) draw_rect( bbox, GRAY, GRAY, 0, H_TPP, 0 );
				draw_rect_mod( bbox, MOD_IN, 0, PH(2) );
				} 
			else draw_rect_mod( bbox, MOD_OUT, 0, PH(2) );
			}	 
		else
			{

//.....Modif for debug the enclosed option by Guy Grimard at Multimeg...........

			if( element->type == TEXT )
				{	 
				buf_printf( "%d %d %d %d enc\n", bbox->w, bbox->h, bbox->x, bbox->y );
				}				 
			else buf_printf( "%d %d %d %d enc\n", bbox->w, bbox->h, bbox->x, bbox->y );

//.................................................................

			}
		}
 	else 
		{
		if( element->attr & SHADOWED ) draw_shadow( element, BLACK );
		}
}

draw_label( label, elmt, baseline )
	LABEL_DATA *label;	
	HDR *elmt;
	unsigned baseline;
{
	FXY_AREA 	elmt_xy;
	char 		string_tmp[50], *string;
	IRECT_AREA 	pos;
	unsigned	font_index;

	if( label == NULL ) return (0);

#if 0
	showsize( elmt );
#endif

	pos.width = pos.height = 0;

	memmove( string_tmp, label->text, label->length );
	string_tmp[label->length] = '\0';
	if( label->options & LABEL_COLON ) strcat( string_tmp, ":" );
	string = strformat( string_tmp );

	if( elmt->type != TEXT && elmt->type != BUTTON ) font_index = 0;
	else font_index = find_font( elmt, NULL, NULL );

	/* _row and _col can be used to override default positioning */


//..........Modif for debug label by Guy Grimard at Multimeg............

	if (label->options==LABEL_CENTER)
		{
		//...........LABEL_CENTER in X:
		pos.col = (int)(elmt->col) + (int)((int)(elmt->width) - strlen(string)*font->map.width)/2;
		}

	else 
		{
		if( elmt->_col )
			pos.col = elmt->_col;
		else
			x_position_label(elmt,label, &pos, &font[font_index], string_tmp);
		}	


	if (label->options==LABEL_CENTER)
		{
		//...........LABEL_CENTER in Y:
		y_position_label( elmt, label, &pos, &font[font_index], baseline );
		}
	else
		{
		if( elmt->_row )
			pos.row = elmt->_row - PV(4);
		else
			y_position_label( elmt, label, &pos, &font[font_index], baseline );
		}

//.......................................................................

	xy_map( &pos, &elmt_xy );
	if ( !label->color ) label->color = BLACK;
buf_printf("(%s) %d %d /QWFont%d %s text\n",string, elmt_xy.x, elmt_xy.y, font_index, ps_colour( label->color ) );
  
	return( 1 );
}

void x_position_label( elmt, label, pos, font, string )
	HDR 		*elmt;
	LABEL_DATA 	*label;
	IRECT_AREA 	*pos;
	union ufont	*font;
	char 		*string;
{
	unsigned short offset;

	if( elmt->attr & ENCLOSED )
		offset = PH(3);

	switch (label->options & (LABEL_LEFT|LABEL_RIGHT|LABEL_ABOVE) ) {
	case LABEL_LEFT:
/*
		pos->col = elmt->col - PH(8) - strlen(string) * font->map.width;
*/
		pos->col = elmt->col - label->offset - strlen(string) * font->map.width - offset;
		break;

	case LABEL_RIGHT:
		pos->col = elmt->col + elmt->width + label->offset + offset;
		break;

	default:
		switch (label->options & (LABEL_LJUST|LABEL_RJUST|LABEL_CENTER|LABEL_OVER) ) {
		case LABEL_LJUST:	
		case LABEL_RJUST:
			pos->col = elmt->col;
			break;
/* doesn't appear to be implemented 
		case LABEL_RJUST:
			pos->col = elmt->col + elmt->width - strlen(string) * font->map.width;
			break;
*/
		case LABEL_OVER:
		case LABEL_CENTER:
			/* Typecasting from hell... */
			pos->col = (int)(elmt->col) + (int)((int)(elmt->width) - strlen(string)*font->map.width)/2;
			break;			
		}	
		break;

	}
}

void y_position_label ( elmt, label, pos, font, baseline )
	HDR 		*elmt;
	LABEL_DATA 	*label;
	IRECT_AREA 	*pos;
	union ufont	*font;
	unsigned	baseline;
{
	unsigned short offset;


	if( elmt->attr & ENCLOSED ) offset = PV(4);

	switch (label->options & (LABEL_LEFT|LABEL_RIGHT|LABEL_ABOVE|LABEL_OVER) ) {
	case LABEL_LEFT:
	case LABEL_RIGHT:
		if ( baseline )
			pos->row = baseline;
		else {
	    	if (label->options & LABEL_CENTER)
    	    	pos->row = elmt->row + ( elmt->height + font->map.height ) /2 - font->map.height + font->map.baseline;
	    	else
    	    	pos->row = elmt->row + font->map.baseline;
		}
	    break;

	case LABEL_OVER:
   	    pos->row = elmt->row + ( elmt->height + font->map.height ) /2 - font->map.height + font->map.baseline;
		break;

	case LABEL_ABOVE:
    	pos->row = elmt->row - label->offset - offset;
	    break;

	/* LABEL_BELOW */
	default:
//    	pos->row = elmt->row + elmt->height + label->offset + font->map.height + offset;
    	pos->row = elmt->row + elmt->height + label->offset + (font->map.height/2) + offset;
	    break;
	}
}

#ifdef __WATCOMC__
STATIC load_font( element, face, style )
	HDR		*element;
	char 	*face;
	int		style;
{
	int ps_height, ps_width;
	char *ps_font, *p;
	FILE *fp;

	if( NULL == ( fp = fopen( "/windows/config/ps.fontmap", "r" ) ) ) {
		flog( "Could not open /windows/config/ps.fontmap\n" );
		return( 0 );
	}

	flog( "Loading font: %s size: %d style: %d\n", face, element->size.text.height, style );

	while( 1 ) {
		if( !fgetline( fp, tbuf, _POSIX_PATH_MAX ) ) {
			flog( "Font \"%s\" not found\n", face );
			break;
		}
		if( *tbuf == '#' )
			continue;
		if( !strncmp( face, tbuf, strlen( face ) ) ) {
			strcpy( font[num_fonts].map.typeface, face );
			StrTok( tbuf ); 
			p = StrTok( NULL );
			font[num_fonts].map.style = _FONT_NORMAL;
			if( strchr( p, 'B' ) )
				font[num_fonts].map.style |= _FONT_BOLD;
			if( strchr( p, 'I' ) )
				font[num_fonts].map.style |= _FONT_ITALIC;
			if( font[num_fonts].map.style == style ) {
				font[num_fonts].map.height = PV( atoi( StrTok( NULL ) ) );
				if( font[num_fonts].map.height == element->size.text.height ) {
					ps_font = StrTok( NULL );
					ps_height = PV( atoi( StrTok( NULL ) ) );
					ps_width = PH( atoi( StrTok( NULL ) ) );
					font[num_fonts].map.baseline = PV( atoi( StrTok( NULL ) ) );
					font[num_fonts].map.width = PH( atoi( StrTok( NULL ) ) );
					font[num_fonts].map.xspacing = (font[num_fonts].map.width - ps_width)/11 * 16;
					buf_printf( "/QWFont%d [%.2f 0 0 %d 0 0] /%s MakeQWFont\n",
						num_fonts, 15.3*(ps_width/H_TPP), -ps_height, ps_font );
					fclose( fp );
					return( num_fonts++ );
				} 
			} 
		}	
	}				
	fclose( fp );
	flog( "No mapping found\n" );
	return( 0 );
}
#endif

find_font( element, text, face )
	HDR 			*element;
	TEXT_ELEMENT	*text;
	char			*face;
{
	int i, style ;

#ifdef __WATCOMC__
/* hacking code to fix a problem for Ross -- some text elements
   seem to have a typeface of PCterm specified */
	if( face && text && strcmp( face, "PCterm" ) && strcmp( face, "System" ) ) {
		flog("Typeface: %s\n",face);
		style = (text->italic?FONT_ITALIC:0) | 
				(element->size.text.weight>=10?_FONT_BOLD:0);
		for( i=3; i < num_fonts; i++ ) {
			if( !strcmp( face, font[i].map.typeface ) && 
				font[i].map.height == element->size.text.height &&
				font[i].map.style == style )
					return( i );
		}
		return( load_font( element, face, style ) );
	} 
#endif

	if( face )
		text->face = 0;

	switch (element->size.text.height/V_TPP) {
	case 14: /* PC Term 15 pt */
		return( 1 );
	case 12: /* System 13 pt */
		return( 2 );
	case 11: /* System 12 pt */
	default:
		return( 0 );
	}
}

void get_points( compact, points0, index, row, col )
	unsigned 	compact;
	POINT 		*points0;
	unsigned 	index, *row, *col;
{
	CPOINT *points1;

	if (compact) {
		points1 = (CPOINT *)points0;
		*row = (points1 + index)->row;
		*col = (points1 + index)->col;
	} else {
		*row = points0[index].row;
		*col = (points0 + index)->col;
	}
}

STATIC set_wind_flag ( opts, opt, on )
	register unsigned *opts, opt;
	bool on;
{
	if (opts)
		if (on)
			*opts |= opt;
		else
			*opts &= ~opt;
}

STATIC set_pane_flag ( opts, opt, on )
	register char *opts, opt;
	bool on;
{
	if (opts)
		if (on)
			*opts |= opt;
		else
			*opts &= ~opt;
}

unsigned get_window_options ( wind, windopts, paneopts )
	struct wind_struct 	*wind;
	char 				*windopts;
	char 				*paneopts[4];
{
	register char *options = windopts;
	int i;
	bool on;

	wind->windopts = WIND_DEFAULTS;

	for( i=0; i<=3; i++ )
		wind->pane[i].opts = PANE_DEFAULTS;

	for( ; *options ; ++options ) {
		if( *options == '-' ) {
			on = NO;
			++options;
		} else
			on = YES;

		switch( *options ) {
			case 'C':	set_wind_flag( &wind->windopts, CENTER_TITLE, on );
						break;
			case 'M':	set_wind_flag( &wind->windopts, MODEL_FRAME, on );	
						break;                                                                                                                                                                                                                                                                                                                                                                                                                                             
			case 'N':	set_wind_flag( &wind->windopts, SUPRESS_TITLE, on );
						break;
			case 'p':	set_wind_flag( &wind->windopts, PIN_FRAME, on );		
						break;                                                                                                                                                                                                                                                                                                                                                                                                                                             
			case 'r':   set_wind_flag( &wind->windopts, RESIZE_TABS, on );	
						break;              	                                                                                                                                                                                                                                                                                                                                                                                                                                   
			case 's':   set_wind_flag( &wind->windopts, SHADOW_FRAME, on );
						break;                                                                                                                                                                                                                                                                                                                                                                                                                                                 
			case 'U':   set_wind_flag( &wind->windopts, ULINE_TITLE, on );	
						break;                                                                                                                                                                                                                                                                                                                                                                                                                                                 
			case 'w':   set_wind_flag( &wind->windopts, BUTTON_FRAME, on );
						break;                                                                                                                                                                                                                                                                                                                                                                                                                                                 
			case 'X':   set_wind_flag( &wind->windopts, FULL_ULINE, on );	
						break;                                                                                                                                                                                                                                                                                                                                                                                                                                                 

		}
	}

	for( i=0; i<=3; i++ ) {
		options = paneopts[i];
		for( ; *options ; ++options ) {
			if( *options == '-' ) {
				on = NO;
				++options;
			}   	
			else    	
				on = YES;
	
			switch( *options ) {
				case 'b':	set_pane_flag( &wind->pane[i].opts, PANE_ASCROLL, on); break;
				case 'H':	set_pane_flag( &wind->pane[i].opts, PANE_HSCROLL, on); break;
				case 'V':	set_pane_flag( &wind->pane[i].opts, PANE_VSCROLL, on); break;
				case 'O':	set_pane_flag( &wind->pane[i].opts, PANE_OUTLINE, on); break;
			}
		}
	}
/*
	flog("Window opts: %b\n",wind->windopts);
*/
	return( wind->windopts );
}

		

/*--------------------------------------------------------
	String manipulation

	strformat:	format a string for PostScript
	trim:		remove padding spaces from a string
----------------------------------------------------------*/

static char strbuf[255];

char *strformat( str )
	char *str;
{
	register unsigned char *p , *p1 ;

#if 0
	zero( tbuf, 101 );
#endif

	for (p = str, p1 = strbuf ; *p ; ++p, ++p1) {
		switch( *p & 0xff ) {
		case '(':
		case ')':
		case 0x5C:
			*p1 = 0x5C;
//			*(++p1) = char_map[*p];
			*(++p1) = *p;
			break;
		/* Frank seems to spit these out once in a while */
		case 0xff:
			*p1 = 0x20;
			break;
		default:
//			if( isprint( char_map[*p] ) ) {
			if( isprint( *p ) ) {
//				*p1 = char_map[*p];
  				*p1 = *p;
  			} else {
				p1 += sprintf( p1, "\\%o", *p ) - 1;
			}
		}
			
	}
	*p1 = '\0';
	return(strbuf);
}

char *trim( str )
    char *str;
{
    int len, ok, sub;

    if (0==(len=strlen(str)))
        return(str);

    ok = 1;
    while (ok) {
        if (*str==0x20) {
            memmove(str, str+1, len);
            len --;
        }
        else
            ok--;
    }

    if (len > 0)
    for (sub=len-1; sub>=0; --sub) {
        if (*(str+sub)==0x20)
            *(str+sub)=0x00;
        else
            sub=0;
    }

	return(str);
}

#ifdef __WATCOMC__
char *StrTok( char *s1 )
#elif !defined __WATCOMC__
char *StrTok( s1 )
	char *s1;
#endif
{
	static char delim[] = { 0x20, '\t', '\0' };

	int go = 1;
	char *p;

	while( go ) {
		p = strtok( s1, delim );
		if( isalnum( *p ) || p == NULL )
			go = 0;
	}
	return( p );
}

char *salloc ( size )
	unsigned size;
{
	char *ptr;

	ptr = malloc ( size );
	if ( !ptr )
		error ( "PICTPS fatal error: malloc failed\n" );
	return ( ptr );
}
		
void showsize ( element )                                                            
    HDR *element;                                                               
{                                                                               
    flog("Tag: %s row %u col %u height %u width %u\n",TAG(element),element->row,element->col,
    element->height,element->width);                                            
}

#ifdef __WATCOMC__
double round ( x )
	double x;
{
	if ( (ceil(x) - x) >= 0.5 )
		return ( ceil(x) );
	else
		return ( floor(x) );
}
#endif

isprintable( str )
    char *str;
{
    register char *p = str;

    for ( ; *p; ++p )
        if ( !isprint( *p ) ) return ( 0 );

    return ( 1 );
}
