/***********************************************************
    PROGRAM:    pictps
    MODULE:     pictps.h

    AUTHOR:     Patrick Hayden
    DATE:       1991-10-29

************************************************************/

#ifdef __WATCOMC__
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mig.h"
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <windows/windows.h>
#include "qwprint.h"
#include "pictps_proto.h"
#elif !defined __WATCOMC__
#include <stdio.h>
#include <math.h>
#include <magic.h>
#include <time.h>
#include <windows.h>
#include "pictps_noproto.h"
#include "qwprint.h"
#endif


#define DEBUG

#define CASE(str)   ( (str)[0]<<8 | (str)[1] )
#define TAG(eptr)	( (char *)eptr + sizeof (HDR) )
#define STATIC		

#define	HERE()		fprintf( stderr, "%s: line %d\n",__FILE__,__LINE )

#ifdef __WATCOMC__
	#define VOID		void
#elif !defined __WATCOMC__
	#define	min(x,y)	(x<y?x:y)
	#define VOID
	#define SEEK_SET 0
	#define memmove memcpy
	#define My_pid	My_tid
	#define	stroke	dash
	#define space	gap
	#define stroke2	dash2
	#define LABEL_OVER		0x8
	#define _POSIX_PATH_MAX	255
	extern char *malloc(), *calloc(), *salloc();
	extern char *strrchr(), *strformat(), *strpbrk(), *strtok();
#endif

/* Where is pi? */
#define PI	3.14159265

/* Limits */
#define	MAX_GROUP_NEST	14
#define MAX_COLOUR		256
#define MAX_FONT		20

/* Number of open link pictures */
#define MAX_LINKS 10

/* Default buffer size */
#define	ELEMENT_BUFF_SIZE	8200

/* Window bars */
#define BTOP 	0
#define BBOTTOM 1
#define BRIGHT 	2
#define BLEFT 	3
#define BPANE 	4

/* Paper sizes 		*/
#define NUM_PAPER_SIZES 4
struct paper_size {
	char 	*name;
	unsigned width, height;
};

/* RGB colour structure */
struct rgb {
	float r, g, b;
};

/* Simplified font struct */

#define	_FONT_NORMAL		0x0000
#define _FONT_ITALIC		0x0001
#define _FONT_BOLD			0x0002

struct sys_font {
	short int		height;
	short int		width;
	short int		baseline;
	char 			*ps_font;
	unsigned short	ps_sx, ps_sy;
};

struct map_font {
	short int		height;
	short int		width;
	short int		baseline;
	short int		xspacing;
	char			style;
	char 			typeface[FACE_LEN];
};

union ufont {
	struct sys_font sys;
	struct map_font map;
};

/* Window info */
struct wind_struct {
	struct bar_offs {
		long top, bottom, right, left;
	} bar;
	struct pane_info {
		char opts;
		RECT_AREA pos;
		RECT_AREA view;
		long picture;
	} pane[4];
	struct state_info {
		unsigned active:1, iconified:1, pinned:1;
	} state;
	long icon;
	char title[120];
	char strstate[40];
	BOX size;
	unsigned windopts;
};

typedef struct fxy_area {
	int x,y;
	int h,w;
} FXY_AREA;

/* Current draw position info */
struct drawing_at {
	bool			active;
	unsigned 		row, col;
	CPOINT			offset;
	CPOINT			spacing;
	RECT_AREA		last;
};

/* Int RECT_AREA */
typedef struct area_dat {
	int row, _row, col, _col, height, _height, width, _width;
} IRECT_AREA;

/* Window display options */
#define CENTER_TITLE	0x0001
#define MODEL_FRAME		0x0002
#define OUTLINE_FRAME	0x0004
#define PIN_FRAME		0x0008
#define RESIZE_TABS		0x0010
#define SHADOW_FRAME	0x0020
#define ULINE_TITLE		0x0040
#define BUTTON_FRAME	0x0080
#define FULL_ULINE		0x0100
#define SUPRESS_TITLE	0x0200

#define PANE_OUTLINE	0x01
#define PANE_ASCROLL	0x02
#define PANE_HSCROLL	0x04
#define PANE_VSCROLL	0x08

#define WIND_DEFAULTS	CENTER_TITLE | MODEL_FRAME | OUTLINE_FRAME | RESIZE_TABS | SHADOW_FRAME | ULINE_TITLE | BUTTON_FRAME
#define PANE_DEFAULTS	PANE_OUTLINE


/* Useful macros for dealing with tips */

#define PV(x)   		(x * V_TPP)
#define PH(x)   		(x * H_TPP)

#define ROUND_PIXELX(x)	( ( ( x+H_TPP/2 ) / H_TPP ) * H_TPP )
#define ROUND_PIXELY(x)	( ( ( x+V_TPP/2 ) / V_TPP ) * V_TPP )

#define	MOD_IN	0
#define MOD_OUT	1

/* Standard glyph sizes */
#define ABBR_H      PV(18)
#define ABBR_W      PH(17)

#define CHECK_H     PV(18)
#define CHECK_W     PH(18)

#define CORNER_H    PV(9)
#define CORNER_W    PH(9)

#define VERT_H      PV(11)
#define VERT_W      PH(11)

#define HORZ_H      PV(13)
#define HORZ_W      PH(11)

#define PINI_H      PV(14)
#define PINI_W      PH(14)

#define PINO_H      PV(12)
#define PINO_W      PH(22)

#define RADIO_H     PV(13)
#define RADIO_W     PH(12)

#define SCROLLV_H   PV(33)
#define SCROLLV_W   PH(15)

#define SCROLLH_H   PV(15)
#define SCROLLH_W   PH(33)

#define END_H       PV(11)
#define END_W       PH(11)

#define ENDH_H      PV(15)
#define ENDH_W      PH(8)

#define ENDV_H      PV(8)
#define ENFV_W      PH(15)

#define MARKSLIDE_H PV(3)
#define MARKSLIDE_W PH(3)

#define MENURIGHT_H	PV(9)
#define MENURIGHT_W	PH(6)

#define MENUDOWN_H	PV(7)
#define MENUDOWN_W	PH(7(


