/***********************************************************
    PROGRAM:    pictps
    MODULE:     ps_driver.c
    AUTHOR:     Patrick Hayden
    DATE:       1991-10-29

    DESCRIPTION:

	PostScript-specific routines for pictps
************************************************************/

#include "pictps.h"
#include "externs.h"

static FXY_AREA     	pic_area;
static XY_POINT			offset;
static float			x_map, y_map;

static XY_POINT			offset_stack[20];
static int				offset_index = -1;
extern int text_number;

/*-------------------------------------------------------
    Import: picture width, picture height
    Defines:    x_map and y_map scaling factors
                pic_area structure
---------------------------------------------------------*/

def_xy_map( x_max, y_max )
    unsigned x_max, y_max;
{

	if ( options & OPT_ONE_TO_ONE ) {
		x_map = 1.0/H_TPP;
		y_map = 1.0/V_TPP;
		pic_area.x = margin.x;
		pic_area.y = margin.y;
		pic_area.w = (int)( x_max * x_map );
    	pic_area.h = paper_sizes[paper].height - margin.h - margin.y;
	} else {
		/* Undefined dimensions */
		if ( !x_max || !y_max)
        	return(0);

	    pic_area.x = margin.x;
   		pic_area.y = margin.y;
    	pic_area.w = paper_sizes[paper].width - margin.w - margin.x;
    	pic_area.h = paper_sizes[paper].height - margin.h - margin.y;
	    x_map = (float)pic_area.w / x_max;
	   	y_map = (float)pic_area.h / y_max;
		x_map = y_map = min( x_map, y_map );
#if 0
		if( x_max > y_max ) {
			pic_area.h = pic_area.h*(float)y_max/x_max;
			pic_area.y = (paper_sizes[paper].height - margin.h )-pic_area.h;
		} else
			pic_area.w = (paper_sizes[paper].width - margin.w - margin.x) *
						(float)x_max/y_max + margin.x;
#else
		pic_area.w = x_map * x_max;
		pic_area.y = margin.y + ( pic_area.h - y_map * y_max );
		pic_area.h = y_map * y_max;
#endif
	}
	clear_offset();
	return( 1 );
}

/*--------------------------------------------------------
	Define origin for points (line drawing)
----------------------------------------------------------*/

void push_offset( VOID )
{
	if( offset_index < 19 ) {
		++offset_index;
		offset_stack[offset_index].x = offset.x;
		offset_stack[offset_index].y = offset.y;
	}
}

void pop_offset( VOID )
{
	if( offset_index > -1 ) {
		offset.x = offset_stack[offset_index].x;
		offset.y = offset_stack[offset_index].y;
		--offset_index;
	}
}
	
void add_offset( area )
	RECT_AREA *area;
{
	offset.x += area->col;
	offset.y += area->row;
}

void sub_offset( area )
	RECT_AREA *area;
{
	offset.x -= area->col;
	offset.y -= area->row;
}

void clear_offset()
{
	offset.x = offset.y = 0;
}

/*-------------------------------------------------
    Maps QNX Windows tips -> PostScript default CTM
	Note: height subtracted from y co-ord
---------------------------------------------------*/

void xy_map( area, xy )
    IRECT_AREA 	*area;
    FXY_AREA 	*xy;
{
	/* Round these to the nearest pixel to avoid alignment problems */
	xy->x = ROUND_PIXELX( area->col + offset.x);
	xy->w = ROUND_PIXELX( area->width );
	xy->y = ROUND_PIXELY( area->row + offset.y );
	xy->h = ROUND_PIXELY( area->height );
}


/*-------------------------------------------------------
	Map a point from TIPS -> Postscript CTM
---------------------------------------------------------*/

void pxy_map( x, y, old_x, old_y )
	int *x, *y;
	int old_x, old_y;
{
	*x = ROUND_PIXELX( offset.x + old_x );
	*y = ROUND_PIXELY( offset.y + old_y );

}

/*--------------------------------------------------------
  Write standard PS-Adobe-3.0 conforming header to outfile
  Write PostScript prolog to outfile
----------------------------------------------------------*/

void ps_write_hdr()                                                                                                
{                                                                                                             
    time_t		timer;                                                                                               
	FILE 		*prolog;
	FXY_AREA 	dim;
	int i;
                                                                                                              
    /* Write standard comments to file */                                                                     
	if ( !( options & OPT_EPS ) )
		buf_printf("%%!PS-Adobe-3.0\n");
	else
		buf_printf("%%!PS-Adobe-3.0 EPSF-3.0\n");

	timer = time ( NULL );

	/* These comments are "strongly recommended" for all PS files */
	if( flags & STAT_TMPFILE )
		buf_printf("%%%%Title: PICTPS file" );
	else
		buf_printf("%%%%Title: %s\n", ifilename );
	buf_printf("%%%%Creator: pictps\n");
	buf_printf("%%%%CreationDate: %s",ctime( &timer ) );

	if ( !( options & OPT_MIN_EPS ) ) {
		buf_printf("%%%%DocumentData: Clean8Bit\n");
		buf_printf("%%%%For: QNX user\n");
		buf_printf("%%%%Orientation: %s\n",(paper % 2)?"Landscape":"Portrait");
		buf_printf("%%%%Pages: 1\n");
		buf_printf("%%%%PageOrder: Ascend\n");
		buf_printf("%%%%Routing: Unknown\n");
		buf_printf("%%%%Version: %.2f %u\n",version,revision);
	}

    def_xy_map( pict_size.width, pict_size.height );                                                                       

	if (paper % 2) {
    	/* Landscape mode */
	    dim.x = pic_area.y;
    	dim.y = pic_area.x;
	    dim.w = pic_area.h + pic_area.y;
    	dim.h = pic_area.w + pic_area.x;
	} else {
    	/* Portrait mode */
	    dim.x = pic_area.x;
    	dim.y = pic_area.y;
    	dim.w = pic_area.w + pic_area.x;
	    dim.h = pic_area.h + pic_area.y;
	}

	buf_printf("%%%%BoundingBox: %u %u %u %u\n",(unsigned)dim.x,
		(unsigned)dim.y,(unsigned)dim.w,(unsigned)dim.h);

	if( !( options & OPT_MIN_EPS ) ) {
		buf_printf("%%%%DocumentNeededResources: font %s\n",font[0].sys.ps_font);
		buf_printf("%%%%+ font %s %s\n",font[1].sys.ps_font, font[2].sys.ps_font);
		buf_printf("%%%%DocumentSuppliedResources: procset pictps_prolog 1.1 0\n");
		buf_printf("%%%%ProofMode: TrustMe\n");
		buf_printf("%%%%EndComments\n\n");

		buf_printf("%%%%BeginProlog\n");
		buf_printf("%%%%BeginResource: procset pictps_prolog\n");
	}

    /* Concatenate standard functions to header */                                                            
	if ((prolog = fopen ("/windows/drivers/qw.prolog","r")) == NULL)
		if ((prolog = fopen ("/windows/config/qw.prolog","r")) == NULL)
			error("Could not open prolog file\n");

	while ( fgetline( prolog, path, 280 ) )
		buf_puts( path );

	fclose( prolog );

	if( !( options & OPT_MIN_EPS ) ) {
		buf_printf( "%%%%EndResource\n" );
		buf_printf( "%%%%EndProlog\n" );
        buf_printf( "%%%%BeginSetup\n" );
	}

	/* Print landscape */
	if ( paper % 2 ) {
		buf_printf("0 %u translate\n",paper_sizes[paper].width );
		buf_printf("-90 rotate\n");
	}

	ps_write_colour_map();
	
	/* Define our fonts */
	for( i=0; i<num_fonts; i++ )
		buf_printf("/QWFont%d [%d 0 0 %d 0 0] /%s MakeQWFont\n",
			i, font[i].sys.ps_sx, font[i].sys.ps_sy, font[i].sys.ps_font );

	/* Now that are values are used, set others */
	font[0].map.xspacing = 0;

	/* Font independent of our scaling */
	buf_printf("/QWFontX [12 0 0 12 0 0] /Courier-Bold MakeQWFont\n");
		
	/* Set page scaling factor */
	buf_printf( "/TPP %u def\n", H_TPP );

	/* Define integer multiple to avoid multiplication in PostScript */
	buf_printf( "/2TPP %u def\n/3TPP %u def\n/4TPP %u def\n/5TPP %u def\n", 
		2 * H_TPP, 3 * H_TPP, 4 * H_TPP, 5 * H_TPP );
	buf_printf( "/6TPP %u def\n/7TPP %u def\n/8TPP %u def\n/9TPP %u def\n",
		6 * H_TPP, 7 * H_TPP, 8 * H_TPP, 9 * H_TPP );
	buf_printf( "/TPPd2 %u def\n", H_TPP /2 );

	/* Scaling factors used for glyphs */
	buf_printf("/PSCLfactor %.2f def\n",1.0/0.122);
	buf_printf("/TIPfactorscale %.2f def\n",PH(2)*0.122);

	if( options & OPT_ANNOTATE )
		Draw_msgs( NULL );

	/* 	Make the CTM map QNX Windows co-ords to TIPS
	 *	x' = ax + cy + t(x)
	 *	y' = bx + dy + t(y)
	 *	b = c = 0
 	 *	
	 * 				 a b c d    t(x) t(y) */
	buf_printf( "[%.4f 0 0 %.4f %u %u] concat\n",
		x_map,	-y_map, pic_area.x, 
		paper_sizes[paper].height - margin.h ); 

	if( !( options & OPT_MIN_EPS ) )
		buf_printf("%%%%EndSetup\n");

#if 0
	if ( !( options & OPT_EPS ) ) {
		buf_puts("%%Page: i 1\n");
		buf_puts("save\n");
	}
#endif
}                                                                                                             

void ps_write_trailer()
{
#if 0
	buf_printf("showpage\n");
#endif

	buf_printf("showpage\nend\nrestore\n");
	if ( !( options & OPT_EPS ) ) {
		buf_printf("%%%%PageTrailer\n");
	}

	if( !( options & OPT_MIN_EPS ) ) {
		buf_printf("%%%%Trailer\n");
		buf_printf("%%%%EOF\n");
	}

	if ( !( options & OPT_EPS ) )
		buf_printf("%c",0x4);
}

char *ps_colour(colour)
int colour;
{
    static char buff[3][27];
	static int i = -1;
	
	/* Multiple buffers are needed */
	i = (i+1)%3;

	if( flags & STAT_DIMMING && colour ) {
		sprintf( buff[i], "[ %.3f %.3f %.3f ]",
			( rgb_map[colour].r + 0.72 ) / 2.0,
			( rgb_map[colour].g + 0.72 ) / 2.0,
			( rgb_map[colour].b + 0.72 ) / 2.0 );
	} else {
	    if( colour <= 16 )
	        sprintf( buff[i], "CC%d", colour );
    	else
	        sprintf( buff[i], "[ %.3f %.3f %.3f ]", rgb_map[colour].r, 
				rgb_map[colour].g, rgb_map[colour].b );
	}
	   
    return(buff[i]);
}

void ps_write_colour_map()
{
	int i;

	for (i=0;i<=16;i++)
		buf_printf("/CC%d [ %.3f %.3f %.3f ] def\n", i, rgb_map[i].r,
		rgb_map[i].g,rgb_map[i].b);
}

void ps_draw_bkgd( picture )
	PICTURE_HDR *picture;
{
	RECT_AREA bkgd;
	FXY_AREA elmt_xy;

	bkgd.row = bkgd.col = 0;
	bkgd.height = pict_size.height;
	bkgd.width  = pict_size.width;

	if ( flags & STAT_WINDOW ) {
		bkgd.row = bkgd.col = -PV(2);
		bkgd.height += PV(4);
		bkgd.width += PH(4);
	}

	xy_map((IRECT_AREA *)&bkgd,&elmt_xy);
	if ( !( options & OPT_FORCE_BKGD ) ) bcol = picture->bkgd;
	
	draw_rect(&elmt_xy,bcol,bcol,0,1,picture->pattern);
}

/*-----------------------------------------------------
	Attach repeat for now: use a PS for loop
-------------------------------------------------------*/

draw_transform_start( )
{
	if( !transform ) return( 0 );

	if( transform->vertical ) {
		buf_printf("0 %d %d { Gs 0 exch Tr\n",
			transform->delta_row, 
			transform->delta_row * transform->vertical );
	}

	if( transform->horizontal ) {
		buf_printf("0 %d %d { Gs 0 Tr\n",
			transform->delta_col, 
			transform->delta_col * transform->horizontal );
	}

	if( transform->angle )
		buf_printf("Gs %.2f rotate\n", transform->angle);

	return( 1 );
}

draw_transform_end ( )
{
	if( !transform ) return( 0 );
	
	if( transform->horizontal )
		buf_printf( "Gr } for\n" );

	if( transform->vertical )
		buf_printf( "Gr } for\n" );

	if( transform->angle )
		buf_printf("Gr\n");

	return( 1 );
}

/* Line styles: dash pattern */

draw_line_style_start( line )
	LINE_ELEMENT *line;
{
	if( !line->stroke || !line->space )
		return( 0 );

	buf_printf( "Gs ");

	if( line->stroke2 )
    	buf_printf("[%u %u %u %u] 0 SDSH\n",
        	line->stroke, line->space, line->stroke2, line->space );
	else if( line->stroke != line->space )
    	buf_printf("[%u %u] 0 SDSH\n", line->stroke, line->space );
	else
    	buf_printf("[%u] 0 SDSH\n",	line->stroke );

	return ( 1 );
}

draw_line_style_end( line )
	LINE_ELEMENT *line;
{
	if( !line->stroke || !line->space )
		return ( 0 );

	buf_printf("Gr\n");

	return( 1 );
}

int rgb_printf ( rgb )
	struct rgb *rgb;
{
	return ( buf_printf("%.3f %.3f %.3f ", rgb->r, rgb->g, rgb->b) );
}

int hdr_printf( join, cap, line, col1, col2, pat)
	unsigned		join ,cap;
	unsigned 		line;
	int 			col1, col2;
	unsigned char 	pat;
{
	if ( !line ) ++line;
	if ( line <= 10 ) line *= 11;

	if( pat )
		flog( "hdr_printf: pattern %d not supported\n", pat );
	
	return ( buf_printf("%d %d %u %s %s ",
		join,cap,line,ps_colour(col1),ps_colour(col2) ));
/*
	if (pat) {
		buf_printf("/chatch1 ");
	}
*/
}

int name_printf( str, pat)
	char 			*str;
	unsigned char 	pat;
{
	char buf[10];

	strcpy( buf, str );

	if( pat ) {
		flog( "name_printf: pattern %d recognized\n", pat );
	}

/*
	if (pat)
		strcat (buf,"p");
*/

	return ( buf_printf("%s\n",buf) );
}

void draw_rect( xy, color, fill, roundness, thickness, pattern)
	FXY_AREA 		*xy;
	unsigned 		color, fill, roundness, thickness;
	unsigned char 	pattern;
{
	int swap;

	if ( !fill && !color ) return;

	if (roundness) 
		{
		if ( xy->w == 0.0 ) xy->w += 1;
		if ( xy->h == 0.0 ) xy->h += 1;

		xy->w += xy->x;
		xy->h += xy->y;

		/* No massive radii */
		if ( roundness > min( xy->w/2 , xy->h/2 ) )
			roundness = min( xy->w/2 , xy->h/2 );

		hdr_printf( 0, 0, thickness, color, fill, pattern );

//.....Modif for debug slider by Guy Grimard at Multimeg...........

		if (xy->y > xy->h)
			{
			swap=xy->y;
			xy->y=xy->h;
			xy->h=swap;
			}
//.................................................................

		buf_printf( "%d %d %d %d %d ",xy->x, xy->y, xy->w, xy->h, roundness);
		name_printf( "rrect", pattern );
		} 
	else 
		{
		hdr_printf( 0, 0, thickness, color, fill, pattern );

//.....Modif for debug rect by Guy Grimard at Multimeg...........

		buf_printf("%d %d %d %d ", xy->w-PV(1), xy->h-PH(1), xy->x, xy->y );

//.................................................................

		name_printf( "rect", pattern );
		}
}

void draw_rect_mod( xy, type, roundness, thickness)
	FXY_AREA 		*xy;
	unsigned		type;
	unsigned 		roundness, thickness;
{
	FXY_AREA xy1, xy2;
	unsigned color;

	if (roundness) {
		memcpy(&xy1,xy,sizeof(FXY_AREA));
		if ( type == MOD_OUT )
			color = DARK_GRAY;
		else
			color = WHITE;
        draw_rect(&xy1,color,TRANSPARENT,roundness,thickness,0);
		memcpy(&xy2,xy,sizeof(FXY_AREA));
		bevelclip_start( &xy2, roundness );
			memcpy(&xy1,xy,sizeof(FXY_AREA));
			if ( type == MOD_OUT )
				color = WHITE;
			else
				color = DARK_GRAY;
			draw_rect( &xy1, color, TRANSPARENT, roundness, thickness, 0);
		clip_end();
	} else {
		buf_printf("%d %d %d %d ", xy->w, xy->h, xy->x, xy->y );
		if ( type == MOD_IN )
			buf_puts( "Imod\n" );
		else
			buf_puts( "Omod\n" );
	}
}

/*---------------------------------------------
	Draw H&V scrollbars with UR corner @ x,y
-----------------------------------------------*/

void draw_hscroll( xy )
	FXY_AREA *xy;
{
	FXY_AREA rect;

	rect.x = xy->x;
	rect.y = xy->y;
	rect.w = PH(5);
	rect.h = SCROLLH_H;
	draw_rect_mod( &rect, MOD_OUT, 0, 1 );

	rect.x = xy->x + PH(7);
	rect.y = xy->y + SCROLLH_H/3;
	rect.h = SCROLLH_H/3;
	rect.w = xy->w - PH(12);
	draw_rect( &rect, GRAY, GRAY, 0, 1, 0 );

	buf_printf("%d %d scrollH\n", xy->x+PH(8), xy->y );

	rect.x = xy->x + xy->w - PH(4);
	rect.y = xy->y;
	rect.w = PH(5);
	rect.h = SCROLLH_H;
	draw_rect_mod(&rect,MOD_OUT,0,1);
}


void draw_vscroll( xy )
	FXY_AREA *xy;
{
	FXY_AREA rect;

	rect.x = xy->x;
	rect.y = xy->y;
	rect.w = SCROLLV_W;
	rect.h = PV(5);
	draw_rect_mod(&rect,MOD_OUT,0,1);

	rect.x = xy->x + SCROLLV_W/3;
	rect.y = xy->y + PV(7);
	rect.h = xy->h - PV(12);
	rect.w = SCROLLV_W/3;
	draw_rect(&rect,GRAY,GRAY,0,1,0);
	
	buf_printf("%d %d scrollV\n",xy->x,rect.y + V_TPP );

	rect.x = xy->x;
	rect.y = xy->y + xy->h - PV(4);
	rect.w = SCROLLV_W;
	rect.h = PV(5);
	draw_rect_mod(&rect,MOD_OUT,0,1);
	
}

void draw_text( tmp_string, xy, color, element, text, fnum )
	char 			*tmp_string;   
	FXY_AREA 		*xy;
	unsigned 		color;
	HDR 			*element;
	TEXT_ELEMENT 	*text;	/* Not always but text options all located 5 bytes from top */
	int 			fnum; 	/* Font index */
{
	int temp;
	char *string;
	/*
	float theta_sin, theta_cos;
	*/

	if( color == TRANSPARENT ) return;

	if( text->down_sub )
		buf_printf("%d %d markD\n",xy->x + font[fnum].map.width * (strlen(tmp_string)+1), xy->y - MENUDOWN_H );

	else if( text->right_sub )
		buf_printf("%d %d markR\n",xy->x + font[fnum].map.width * (strlen(tmp_string)+1), xy->y - MENURIGHT_H );

	else if( text->right_mark )
		buf_printf("%d %d mark%d\n",xy->x + font[fnum].map.width * (strlen(tmp_string)+1), xy->y - MENURIGHT_H, 
			text->mark_type );

	if( text->left_mark )
		buf_printf("%d %d mark%d\n",xy->x - 2*font[fnum].map.width, xy->y -MENURIGHT_H , 
			text->mark_type );

	if (text->underline) 
		{
		if (element->attr & MODELLED) temp = 1;
		else temp = 0;
		buf_printf("%d %d %d %d uline\n",temp,xy->w,xy->x, xy->y + V_TPP/2);
		}

//.....Modif for debug text by Guy Grimard at Multimeg...........

	if (text->centered) xy->x=xy->x+(element->width/2)-(font[fnum].map.width*(strlen(tmp_string)))/2;

//.................................................................

	string = strformat(tmp_string);


//.....Modif for debug text by Guy Grimard at Multimeg...........

	if(text_number)
		{
		if( text->face ) 
			{
			buf_printf( "%d (%s) %d %d /QWFont%d %s atext\n", font[fnum].map.xspacing,
			string, xy->x, xy->y, fnum, ps_colour(color) );
			}    														
		else buf_printf( "(%s) %d %d /QWFont%d %s text\n", string, xy->x, xy->y - (2 * (font[fnum].map.height - font[fnum].map.baseline)), fnum, ps_colour(color) );
		}

	else
		{
		if( text->face ) 
			{
			buf_printf( "%d (%s) %d %d /QWFont%d %s atext\n", font[fnum].map.xspacing,
			string, xy->x, xy->y, fnum, ps_colour(color) );
			}    														
 		else buf_printf( "(%s) %d %d /QWFont%d %s text\n", string, xy->x, xy->y, fnum, ps_colour(color) );
		}

//.................................................................


	if (text->strike_thru)
		buf_printf("%d %d %%2.f uline\n",xy->w,xy->x, xy->y + font[fnum].map.baseline/3.0 );
}

/* Set clipping area to box bounded by FXY_AREA *xy */

void boxclip_start( xy )
	FXY_AREA *xy;
{
	buf_printf("Gs %d %d %d %d boxclip\n", xy->w,xy->h,xy->x,xy->y);
}

/* Upper-left half of a rectangle with 45 degree line caps */

void bevelclip_start( xy, width )
	FXY_AREA *xy;
	unsigned  width;
{
	buf_printf("Gs %u %d %d %d %d bevelclip\n",
		width, xy->w,xy->h,xy->x,xy->y);                        	
}

/* Restore old clipping area */

void clip_end()
{
	buf_puts("Gr\n");
}

                  