/***********************************************************
    PROGRAM:    pictps
    MODULE:     win_funs.c

    AUTHOR:     Patrick Hayden
    DATE:       1991-10-29

    DESCRIPTION:

	All functions that interact with screen are kept in this
	file.  Currently, for .pict files, pictps uses screen
	to determine the dimensions of the picture.
************************************************************/

#include "pictps.h"
#include "externs.h"

open_graphics( node )
	unsigned node;
{
	char buf[5], *p;

	if ( flags & STAT_GRAPHICS ) GraphicsClose( 0 );
	
	if (node) {
		sprintf(buf,"[%u]",node);
		p = buf;
	} else 
		p = NULL;

	if( p )
		flog( "Attempting to connect to screen: %s\n", buf );
	else
		flog( "Attempting to connect to local screen\n" );

	if ( GraphicsOpen( p ) ) 
		flags |= STAT_GRAPHICS;
	else 
		error( "Could not connect to screen\n" );

	return ( 1 );
}

int query_map_rgb( VOID )
{
	int i;
	unsigned red, green, blue;
	unsigned colours;

	if ( !( flags & STAT_GRAPHICS ) ) {
		flog("query failed: graphics not opened\n");
		return ( NO );
	}

	rgb_map[0].r = rgb_map[0].g = rgb_map[0].b = -1.0;
	ScreenInfo(NULL,&colours,NULL,NULL,NULL,NULL,NULL,NULL);
	for (i=1;i<=colours;i++) {
		ScreenColorInfo(i,NULL,&red,&green,&blue,NULL);
		if( red > 9500 ) red = 10000;
		if( green > 9500 ) green = 10000;
		if( blue > 9500 ) blue = 10000;
		rgb_map[i].r = red/10000.0;
		rgb_map[i].g = green/10000.0;
		rgb_map[i].b = blue/10000.0;
	}
	return ( YES );
}


int query_font_info( VOID )
{
	REAL_EXTENT size;
	int baseline;
	int id[3];
	int i;

	if ( !( flags & STAT_GRAPHICS ) )
		return ( NO );

	id[0] = 0;
	id[1] = ScreenFont("PCterm",NULL,0,150,NULL);
	id[2] = ScreenFont("PCterm",NULL,0,130,NULL);

	for( i=0; i<=2; i++ ) {
		ScreenFontInfo(id[i],NULL,NULL,&size, &baseline,NULL,NULL);
		font[i].sys.width = (size.w * H_TPP) / 8;
		font[i].sys.height = (size.h * V_TPP) / 8;
		font[i].sys.baseline = (baseline * V_TPP) / 8;
	}
#if 0
	for (i=0;i<3;i++) {	
		flog("x y h w: %u %u %d %d\n",font[i].size.x,font[i].size.y,
			font[i].size.h,font[i].size.w);
		flog("bline: %d\n",font[i].baseline);
	}
#endif
	return ( YES );
}

int query_pict_size( fname )
	char *fname;
{
	POINT	last_point;

	if( !( flags & STAT_GRAPHICS ) )
		return( 0 );

	if ( (! pict_size.height ) || (! pict_size.width ) ) {
		if ( !( pict_id = PictureOpen( "ps",fname,NULL,0,0,NULL,"LR" ) ) )
			error( "Could not open picture: %s\n", fname );
		PictureInfo( NULL,NULL,NULL,&last_point,NULL,NULL );
		pict_size.height = last_point.row + 100;
		pict_size.width = last_point.col + 100;
		flog ("Picture height: %u width: %u\n",pict_size.height,pict_size.width);
	}
	return( 1 );
}

